"""
    This standalone module is originally located at "biothings/standalone" repo.
    It's used for Standalone/Autohub instance.
"""
import os
import logging
import sys
import asyncio
from functools import partial

from biothings import config as btconfig
from biothings.hub import HubServer
from biothings.hub.autoupdate import BiothingsDumper, BiothingsUploader
from biothings.utils.es import ESIndexer
from biothings.utils.backend import DocESBackend
from biothings.utils.loggers import get_logger


class AutoHubServer(HubServer):

    DEFAULT_FEATURES = ["job", "autohub", "terminal", "config", "ws"]


class AutoHubFeature(object):

    DEFAULT_DUMPER_CLASS = BiothingsDumper
    DEFAULT_UPLOADER_CLASS = BiothingsUploader

    def __init__(self, managers, version_urls, indexer_factory=None, *args, **kwargs):
        """
        version_urls is a list of URLs pointing to versions.json file. The name
        of the data release is taken from the URL (http://...s3.amazon.com/<the_name>/versions.json)
        unless specified as a dict: {"name" : "custom_name", "url" : "http://..."}

        If indexer_factory is passed, it'll be used to create indexer used to dump/check versions
        currently installed on ES, restore snapshot, index, etc... A indexer_factory is typically
        used to generate indexer dynamically (ES host, index name, etc...) according to URLs for
        instance. See standalone.hub.DynamicIndexerFactory class for an example. It is typically
        used when lots of data releases are being managed by the Hub (so no need to manually update
        STANDALONE_CONFIG parameter.

        If indexer_factory is None, a config param named STANDALONE_CONFIG is used,
        format is the following:

            {"_default" : {"es_host": "...", "index": "...", "doc_type" : "..."},
             "the_name" : {"es_host": "...", "index": "...", "doc_type" : "..."}}

        When a data release named (from URL) matches an entry, it's used to configured
        which ES backend to target, otherwise the default one is used.
        """
        super().__init__(*args, **kwargs)
        self.version_urls = self.extract(version_urls)
        self.indexer_factory = indexer_factory
        self.managers = managers
        self.logger, _ = get_logger("autohub")

    def extract(self, urls):
        vurls = []
        for url in urls:
            if isinstance(url, dict):
                assert "name" in url and "url" in url
                vurls.append(url)
            else:
                vurls.append({"name": self.get_folder_name(url), "url": url})

        return vurls

    def install(self, src_name, version="latest", dry=False, force=False):
        """
        Update hub's data up to the given version (default is latest available),
        using full and incremental updates to get up to that given version (if possible).
        """
        @asyncio.coroutine
        def do(version):
            try:
                dklass = self.managers["dump_manager"][src_name][0]   # only one dumper allowed / source
                dobj = self.managers["dump_manager"].create_instance(dklass)
                update_path = dobj.find_update_path(version, backend_version=dobj.target_backend.version)
                version_path = [v["build_version"] for v in update_path]
                if not version_path:
                    logging.info("No update path found")
                    return

                logging.info(
                    "Found path for updating from version '%s' to version '%s': %s",
                    dobj.target_backend.version, version, version_path)
                if dry:
                    return version_path

                for step_version in version_path:
                    logging.info("Downloading data for version '%s'", step_version)
                    jobs = self.managers["dump_manager"].dump_src(src_name, version=step_version, force=force)
                    download = asyncio.gather(*jobs)
                    res = yield from download
                    assert len(res) == 1
                    if res[0] is None:
                        # download ready, now install
                        logging.info("Updating backend to version '%s'", step_version)
                        jobs = self.managers["upload_manager"].upload_src(src_name)
                        upload = asyncio.gather(*jobs)
                        res = yield from upload

            except Exception:
                self.logger.exception("data install failed")
                raise
        return asyncio.ensure_future(do(version))

    def get_folder_name(self, url):
        return os.path.basename(os.path.dirname(url))

    def get_class_name(self, folder):
        """Return class-compliant name from a folder name"""
        return folder.replace(".", "_").replace("-", "_")

    def list_biothings(self):
        return self.version_urls

    def configure(self):
        """
        Either configure autohub from static definition (STANDALONE_CONFIG) where
        different hard-coded names of indexes can be managed on different ES server,
        *or* use a indexer factory where index names are taken from version_urls *but*
        only one ES host is used.
        """
        default_standalone_conf = getattr(btconfig, "STANDALONE_CONFIG", {}).get("_default")
        if not default_standalone_conf:
            assert self.indexer_factory, "No STANDALONE_CONFIG defined, and no indexer factory class defined as well"
        for info in self.version_urls:
            version_url = info["url"]
            self.__class__.DEFAULT_DUMPER_CLASS.VERSION_URL = version_url
            if self.indexer_factory:
                pidxr, actual_conf = self.indexer_factory.create(info["name"])
                self.logger.info("Autohub configured for %s (dynamic): %s" % (info["name"], actual_conf))
            else:
                actual_conf = btconfig.STANDALONE_CONFIG.get(info["name"], default_standalone_conf)
                assert actual_conf, "No standalone config could be found for data release '%s'" % info["name"]
                self.logger.info("Autohub configured for %s (static): %s" % (info["name"], actual_conf))
                pidxr = partial(ESIndexer, index=actual_conf["index"],
                                doc_type=None,
                                es_host=actual_conf["es_host"])
            partial_backend = partial(DocESBackend, pidxr)

            SRC_NAME = info["name"]
            dump_class_name = "%sDumper" % self.get_class_name(SRC_NAME)
            # dumper
            dumper_klass = type(
                dump_class_name, (self.__class__.DEFAULT_DUMPER_CLASS,),
                {
                    "TARGET_BACKEND": partial_backend,
                    "SRC_NAME": SRC_NAME,
                    "SRC_ROOT_FOLDER": os.path.join(btconfig.DATA_ARCHIVE_ROOT, SRC_NAME),
                    "VERSION_URL": version_url,
                    "AWS_ACCESS_KEY_ID": btconfig.STANDALONE_AWS_CREDENTIALS.get("AWS_ACCESS_KEY_ID"),
                    "AWS_SECRET_ACCESS_KEY": btconfig.STANDALONE_AWS_CREDENTIALS.get("AWS_SECRET_ACCESS_KEY")
                }
            )
            sys.modules["biothings.hub.standalone"].__dict__[dump_class_name] = dumper_klass
            self.managers["dump_manager"].register_classes([dumper_klass])
            # uploader
            # syncer will work on index used in web part
            esb = (actual_conf["es_host"], actual_conf["index"], None)
            partial_syncer = partial(self.managers["sync_manager"].sync, "es", target_backend=esb)
            # manually register biothings source uploader
            # this uploader will use dumped data to update an ES index
            uploader_class_name = "%sUploader" % self.get_class_name(SRC_NAME)
            uploader_klass = type(
                uploader_class_name, (self.__class__.DEFAULT_UPLOADER_CLASS,),
                {
                    "TARGET_BACKEND": partial_backend,
                    "SYNCER_FUNC": partial_syncer,
                    "AUTO_PURGE_INDEX": True,   # because we believe
                    "name": SRC_NAME
                }
            )
            sys.modules["biothings.hub.standalone"].__dict__[uploader_class_name] = uploader_klass
            self.managers["upload_manager"].register_classes([uploader_klass])



import os
import glob
from subprocess import check_output
from subprocess import CalledProcessError
from setuptools import setup, find_packages

setup_path = os.path.dirname(__file__)


def read(fname):
    return open(os.path.join(setup_path, fname), encoding="utf8").read()


REPO_URL = "https://github.com/biothings/biothings.api"


# get version
version = __import__('biothings').get_version()

# should fail if installed from source or from pypi,
# version gets set to MAJOR.MINOR.# commits on master branch if installed from pip repo
# otherwise to MAJOR.MINOR.MICRO as defined in biothings.version
try:
    num_commits = check_output("git rev-list --count master", shell=True).strip().decode('utf-8')
except CalledProcessError:
    num_commits = ''

# Calculate commit hash, should fail if installed from source or from pypi
try:
    commit_hash = check_output("git rev-parse HEAD", shell=True).strip().decode('utf-8')
except CalledProcessError:
    commit_hash = ''

# Write commit to file inside package, that can be read later
if commit_hash or num_commits:
    with open('biothings/.git-info', 'w') as git_file:
        git_file.write("{}.git\n{}\n{}".format(REPO_URL, commit_hash, num_commits))


# very minimal requirement for running biothings.web
install_requires = [
    'requests>=2.21.0',
    'tornado==5.1.1',
    'gitpython>=3.1.0'
    'elasticsearch>=6, <8',
    'elasticsearch-dsl>=6, <8',
    'elasticsearch-async>=6.2.0'
]

# extra requirements for biothings.web
web_extra_requires = [
    'msgpack>=0.6.1',   # support format=msgpack
    'PyYAML>=5.1',      # support format=yaml
]

# extra requirements to run biothings.hub
hub_requires = [
    'beautifulsoup4',   # used in dumper.GoogleDriveDumper
    'aiocron',          # setup scheduled jobs
    'asyncssh==1.7.1',  # needs libffi-dev installed (apt-get)
    'pymongo',
    'psutil',
    'jsonpointer',      # for utils.jsonpatch
    'IPython',          # for interactive hub console
    'boto',
    'boto3',
    'multiprocessing_on_dill',  # can replace pickler in concurrent.futures
    'dill',             # a pickle alternative with extra object type support
    'pyinotify',        # hub reloader
    'prettytable',      # diff report renderer
    'sockjs-tornado==1.0.6',   # websocket server for HubServer
    'jsonschema>=2.6.0',
    'pip',              # auto-install requirements from plugins
    'pandas==1.0.1',    # json with inf/nan and more to come
    'yapf',             # code reformatter, better results than autopep8
    'requests-aws4auth',    # aws s3 auth requests for autohub
    'networkx>=2.1',            # datatransform
    'biothings_client==0.2.1'   # datatransform (api client)
]

# extra requirements for building docs
docs_requires = [
    'sphinx>=2.4.3',
    'sphinx_rtd_theme>=0.4.3'
]

# extra requirements for nosetests
test_requires = [
    "nose>=1.3.7",
    "pytest"
]

setup(
    name="biothings",
    version=version,
    author="Cyrus Afrasiabi, Sebastien Lelong, Xinghua Zhou, Chunlei Wu",
    author_email="cwu@scripps.edu",
    description="a toolkit for building high-performance data APIs in biology",
    license="Apache License, Version 2.0",
    keywords="biology annotation web service client api",
    url=REPO_URL,
    packages=find_packages(),
    package_data={'': ['*.html']},
    include_package_data=True,
    scripts=list(glob.glob('biothings/bin/*')),
    long_description=read('README.md'),
    long_description_content_type='text/markdown',
    classifiers=[
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Development Status :: 4 - Beta",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
        "Operating System :: POSIX",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "Topic :: Utilities",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
    ],
    install_requires=install_requires,
    extras_require={
        'web_extra': web_extra_requires,
        'hub': hub_requires + test_requires,
        'dev': web_extra_requires + hub_requires + docs_requires + test_requires
    },
)




import json
import logging
import os
from copy import deepcopy

from elasticsearch_dsl import Q


class ESUserQuery():

    def __init__(self, path):

        self._queries = {}
        self._filters = {}
        try:
            for (dirpath, dirnames, filenames) in os.walk(path):
                if dirnames:
                    self.logger.info("Detected user query folders: %s.", dirnames)
                    continue
                for filename in filenames:
                    with open(os.path.join(dirpath, filename)) as text_file:
                        if 'query' in filename:
                            ## alternative implementation
                            # self._queries[os.path.basename(dirpath)] = text_file.read()
                            ##
                            self._queries[os.path.basename(dirpath)] = json.load(text_file)
                        elif 'filter' in filename:
                            self._filters[os.path.basename(dirpath)] = json.load(text_file)
        except Exception:
            self.logger.exception('Error loading user queries.')

    def has_query(self, named_query):

        return named_query in self._queries

    def has_filter(self, named_query):

        return named_query in self._filters

    def get_query(self, named_query, **kwargs):

        def in_place_sub(dic, kwargs):
            for key in dic:
                if isinstance(dic[key], dict):
                    in_place_sub(dic[key], kwargs)
                elif isinstance(dic[key], list):
                    for item in dic[key]:
                        in_place_sub(item, kwargs)
                elif isinstance(dic[key], str):
                    dic[key] = dic[key].format(**kwargs).format(**kwargs)  # {{q}}

        dic = deepcopy(self._queries.get(named_query))
        in_place_sub(dic, kwargs)
        key, val = next(iter(dic.items()))
        return Q(key, **val)

        ## alternative implementation
        # string = self._queries.get(named_query)
        # string1 = re.sub(r"\}", "}}", string)
        # string2 = re.sub(r"\{", "{{", string1)
        # string3 = re.sub(r'\{\{\{\{(?P<var>.*?)\}\}\}\}', r'{\g<var>}', string2)
        # return string3
        ##

    def get_filter(self, named_query):

        dic = self._filters.get(named_query)
        key, val = next(iter(dic.items()))
        return Q(key, **val)

    @property
    def logger(self):
        return logging.getLogger(__name__)



import logging
import cgi
import os

from tornado.web import stream_request_body
from tornado import gen

from .base import GenericHandler
from biothings.utils.common import sizeof_fmt


@stream_request_body
class UploadHandler(GenericHandler):
    def initialize(self, upload_root, **kwargs):
        self.upload_root = upload_root

    def prepare(self):
        # sanity check + extract boundary
        ct = self.request.headers.get("Content-Type")
        if not ct:
            self.write_error(
                400, exc_info=[None, "No Content-Type header found", None])
            return
        try:
            ct, boundary = ct.split(";")
        except Exception as e:
            self.write_error(400, exc_info=[None, str(e), None])
            return
        ct = ct.strip()
        if ct != "multipart/form-data":
            self.write_error(
                400,
                exc_info=[
                    None,
                    "Excepting 'Content-Type: multipart/form-data', got %s" %
                    ct, None
                ])
            return
        boundname, boundary = boundary.strip().split("=")
        if boundname != "boundary":
            self.write_error(
                400,
                exc_info=[None, "No boundary field found in headers", None])
        self.boundary = boundary
        cl = self.request.headers.get("Content-Length")
        if cl:
            self.contentlength = sizeof_fmt(int(cl))
        else:
            self.contentlength = "unknown size"

        self.begin = False
        self.src_name = self.request.uri.split("/")[-1]
        self.fp = None
        self.head = ""

    def parse_head(self):
        data = self.head.splitlines()
        assert "--" + self.boundary == data[0]
        for dat in data[1:]:
            if "Content-Disposition:".lower() in dat.lower():
                cd, cdopts = cgi.parse_header(dat)
                try:
                    self.filename = cdopts["filename"]
                except KeyError:
                    self.write_error(
                        400,
                        exc_info=[
                            None, "No 'filename' found in Content-Disposition",
                            None
                        ])
            if "Content-Type:".lower() in dat.lower():
                _, ct = map(str.strip, dat.split(":"))
                if not ct == "application/octet-stream":
                    self.write_error(
                        400,
                        exc_info=[
                            None,
                            "Expected 'application/octet-stream', got: %s" %
                            ct, None
                        ])
        logging.info("Upload file '%s' (%s) for source '%s'" %
                     (self.filename, self.contentlength, self.src_name))
        folder = os.path.join(self.upload_root, self.src_name)
        os.makedirs(folder, exist_ok=True)
        self.fp = open(os.path.join(folder, self.filename), "wb")

    @gen.coroutine
    def data_received(self, chunk):
        while True:
            if self.begin:
                endbound = b"\r\n--" + self.boundary.encode() + b"--\r\n"
                if endbound in chunk:
                    chunk = chunk.replace(endbound, b"")
                self.fp.write(chunk)
                break
            else:
                # we're in headers, assuming str conversion
                data = chunk.split(b"\r\n\r\n")
                assert len(data) >= 1
                self.head += data[0].decode()
                self.begin = True
                chunk = b"\r\n\r\n".join(data[1:])
                self.parse_head()

    def post(self, src_name):
        assert src_name == self.src_name
        self.write('ok')



'''
This module is used to automatically update the data in a standalone API instance.
 (e.g. used in biothings.standalone, now hub.standalone module, and biothings.farms repo)
'''

from .dumper import BiothingsDumper  # noqa: F401
from .uploader import BiothingsUploader  # noqa: F401



import time, logging, os, io, glob, datetime
import dateutil.parser as dtparser
from functools import wraps
from pymongo import MongoClient, DESCENDING
from pymongo.collection import Collection
from functools import partial
from collections import defaultdict
import bson

from biothings.utils.common import timesofar, get_random_string, iter_n, \
                                   open_compressed_file, get_compressed_outfile, \
                                   dotdict
from biothings.utils.backend import DocESBackend, DocMongoBackend
from biothings.utils.hub_db import IDatabase, ChangeWatcher
# stub, until set to real config module
config = None


class DummyCollection(dotdict):
    def count(self):
        return None
    def drop(self):
        pass
    def __getitem__(self,what):
        return DummyCollection() # ???


class DummyDatabase(dotdict):
    def collection_names(self):
        return []
    def __getitem__(self,what):
        return DummyCollection()


class Database(MongoClient,IDatabase):

    def __init__(self,dbname,*args,**kwargs):
        super(Database,self).__init__(dbname)
        self.name = dbname

def requires_config(func):
    @wraps(func)
    def func_wrapper(*args,**kwargs):
        global config
        if not config:
            try:
                from biothings import config as config_mod
                config = config_mod
            except ImportError:
                raise Exception("call biothings.config_for_app() first")
        return func(*args,**kwargs)
    return func_wrapper

@requires_config
def get_conn(server, port):
    try:
        if config.DATA_SRC_SERVER_USERNAME and config.DATA_SRC_SERVER_PASSWORD:
            uri = "mongodb://{}:{}@{}:{}".format(config.DATA_SRC_SERVER_USERNAME,
                                                 config.DATA_SRC_SERVER_PASSWORD,
                                                 server, port)
        else:
            uri = "mongodb://{}:{}".format(server, port)
        conn = Database(uri)
        return conn
    except (AttributeError,ValueError) as e:
        # missing config variables (or invalid), we'll pretend it's a dummy access to mongo
        # (dummy here means there really shouldn't be any call to get_conn()
        # but mongo is too much tied to the code and needs more work to 
        # unlink it
        return DummyDatabase()


@requires_config
def get_hub_db_conn():
    conn = Database(config.HUB_DB_BACKEND["uri"])
    return conn

@requires_config
def get_src_conn():
    return get_conn(config.DATA_SRC_SERVER, getattr(config,"DATA_SRC_PORT",27017))

@requires_config
def get_src_db(conn=None):
    conn = conn or get_src_conn()
    return conn[config.DATA_SRC_DATABASE]

@requires_config
def get_src_master(conn=None):
    conn = conn or get_hub_db_conn()
    return conn[config.DATA_HUB_DB_DATABASE][config.DATA_SRC_MASTER_COLLECTION]

@requires_config
def get_src_dump(conn=None):
    conn = conn or get_hub_db_conn()
    return conn[config.DATA_HUB_DB_DATABASE][getattr(config,"DATA_SRC_DUMP_COLLECTION","src_dump")]

@requires_config
def get_src_build(conn=None):
    conn = conn or get_hub_db_conn()
    return conn[config.DATA_HUB_DB_DATABASE][config.DATA_SRC_BUILD_COLLECTION]

@requires_config
def get_src_build_config(conn=None):
    conn = conn or get_hub_db_conn()
    return conn[config.DATA_HUB_DB_DATABASE][config.DATA_SRC_BUILD_COLLECTION + "_config"]

@requires_config
def get_data_plugin(conn=None):
    conn = conn or get_hub_db_conn()
    return conn[config.DATA_HUB_DB_DATABASE][config.DATA_PLUGIN_COLLECTION]

@requires_config
def get_api(conn=None):
    conn = conn or get_hub_db_conn()
    return conn[config.DATA_HUB_DB_DATABASE][config.API_COLLECTION]

@requires_config
def get_cmd(conn=None):
    conn = conn or get_hub_db_conn()
    return conn[config.DATA_HUB_DB_DATABASE][config.CMD_COLLECTION]

@requires_config
def get_event(conn=None):
    conn = conn or get_hub_db_conn()
    return conn[config.DATA_HUB_DB_DATABASE][getattr(config,"EVENT_COLLECTION","event")]

@requires_config
def get_hub_config(conn=None):
    conn = conn or get_hub_db_conn()
    return conn[config.DATA_HUB_DB_DATABASE][getattr(config,"HUB_CONFIG_COLLECTION","hub_config")]

@requires_config
def get_last_command(conn=None):
    cmd = get_cmd(conn)
    cur = cmd.find({},{"_id":1}).sort("_id",DESCENDING).limit(1)
    return next(cur)

@requires_config
def get_target_conn():
    if config.DATA_TARGET_SERVER_USERNAME and config.DATA_TARGET_SERVER_PASSWORD:
        uri = "mongodb://{}:{}@{}:{}".format(config.DATA_TARGET_SERVER_USERNAME,
                                             config.DATA_TARGET_SERVER_PASSWORD,
                                             config.DATA_TARGET_SERVER,
                                             config.DATA_TARGET_PORT)
    else:
        uri = "mongodb://{}:{}".format(config.DATA_TARGET_SERVER,config.DATA_TARGET_PORT)
    conn = Database(uri)
    return conn


@requires_config
def get_target_db(conn=None):
    conn = conn or get_target_conn()
    return conn[config.DATA_TARGET_DATABASE]


@requires_config
def get_target_master(conn=None):
    conn = conn or get_target_conn()
    return conn[config.DATA_TARGET_DATABASE][config.DATA_TARGET_MASTER_COLLECTION]

@requires_config
def get_source_fullname(col_name):
    """
    Assuming col_name is a collection created from an upload process,
    find the main source & sub_source associated.
    """
    src_dump = get_src_dump()
    # "sources" in config is a list a collection names. src_dump _id is the name of the
    # resource but can have sub-resources with different collection names. We need
    # to query inner keys upload.job.*.step, which always contains the collection name  
    info = src_dump.find_one({"$where":"function() {if(this.upload) {for(var index in this.upload.jobs) {if(this.upload.jobs[index].step == \"%s\") return this;}}}" % col_name})
    if info:
        name = info["_id"]
        if name != col_name:
            # col_name was a sub-source name
            return "%s.%s" % (name,col_name)
        else:
            return name

def get_source_fullnames(col_names):
    main_sources = set()
    for col_name in col_names:
        main_source = get_source_fullname(col_name)
        if main_source:
            main_sources.add(main_source)
    return list(main_sources)

def doc_feeder(collection, step=1000, s=None, e=None, inbatch=False, query=None, batch_callback=None,
               fields=None, logger=logging):
    '''A iterator for returning docs in a collection, with batch query.
       additional filter query can be passed via "query", e.g.,
       doc_feeder(collection, query={'taxid': {'$in': [9606, 10090, 10116]}})
       batch_callback is a callback function as fn(cnt, t), called after every batch
       fields is optional parameter passed to find to restrict fields to return.
    '''
    if isinstance(collection,DocMongoBackend):
        collection = collection.target_collection
    cur = collection.find(query, no_cursor_timeout=True, projection=fields)
    n = cur.count()
    s = s or 0
    e = e or n
    ##logger.info('Retrieving %d documents from database "%s".' % (n, collection.name))
    t0 = time.time()
    if inbatch:
        doc_li = []
    cnt = 0
    t1 = time.time()
    try:
        if s:
            cur.skip(s)
            cnt = s
            ##logger.info("Skipping %d documents." % s)
        if e:
            cur.limit(e - (s or 0))
        cur.batch_size(step)
        ##logger.info("Processing %d-%d documents..." % (cnt + 1, min(cnt + step, e)))
        for doc in cur:
            if inbatch:
                doc_li.append(doc)
            else:
                yield doc
            cnt += 1
            if cnt % step == 0:
                if inbatch:
                    yield doc_li
                    doc_li = []
                if n:
                    pass
                    ##logger.info('Done.[%.1f%%,%s]' % (cnt * 100. / n, timesofar(t1)))
                else:
                    pass
                    ##logger.info('Nothing to do...')
                if batch_callback:
                    batch_callback(cnt, time.time()-t1)
                if cnt < e:
                    t1 = time.time()
                    ##logger.info("Processing %d-%d documents..." % (cnt + 1, min(cnt + step, e)))
        if inbatch and doc_li:
            #Important: need to yield the last batch here
            yield doc_li

        #print 'Done.[%s]' % timesofar(t1)
        if n:
            pass
            ##logger.info('Done.[%.1f%%,%s]' % (cnt * 100. / n, timesofar(t1)))
        else:
            pass
            ##logger.info('Nothing to do...')
        ##logger.info("=" * 20)
        ##logger.info('Finished.[total time: %s]' % timesofar(t0))
    finally:
        cur.close()


def get_cache_filename(col_name):
    cache_folder = getattr(config,"CACHE_FOLDER",None)
    if not cache_folder:
        return # we don't even use cache, forget it
    cache_format = getattr(config,"CACHE_FORMAT",None)
    cache_file = os.path.join(config.CACHE_FOLDER,col_name)
    cache_file = cache_format and (cache_file + ".%s" % cache_format) or cache_file
    return cache_file


def invalidate_cache(col_name,col_type="src"):
    if col_type == "src":
        src_dump = get_src_dump()
        if not "." in col_name:
            fullname = get_source_fullname(col_name)
        assert fullname, "Can't resolve source '%s' (does it exist ?)" % col_name

        main,sub = fullname.split(".")
        doc = src_dump.find_one({"_id":main})
        assert doc, "No such source '%s'" % main
        assert doc.get("upload",{}).get("jobs",{}).get(sub), "No such sub-source '%s'" % sub
        # this will make the cache too old
        doc["upload"]["jobs"][sub]["started_at"] = datetime.datetime.now()
        src_dump.update_one({"_id":main},{"$set" : {"upload.jobs.%s.started_at" % sub:datetime.datetime.now()}})
    elif col_type == "target":
        # just delete the cache file
        cache_file = get_cache_filename(col_name)
        if cache_file:
            try:
                os.remove(cache_file)
            except FileNotFoundError:
                pass


# TODO: this func deals with different backend, should not be in bt.utils.mongo
# and doc_feeder should do the same as this function regarding backend support
@requires_config
def id_feeder(col, batch_size=1000, build_cache=True, logger=logging,
              force_use=False, force_build=False, validate_only=False):
    """Return an iterator for all _ids in collection "col"
       Search for a valid cache file if available, if not
       return a doc_feeder for that collection. Valid cache is
       a cache file that is newer than the collection.
       "db" can be "target" or "src".
       "build_cache" True will build a cache file as _ids are fetched, 
       if no cache file was found
       "force_use" True will use any existing cache file and won't check whether
       it's valid of not.
       "force_build" True will build a new cache even if current one exists
       and is valid.
       "validate_only" will directly return [] if the cache is valid (convenient
       way to check if the cache is valid)
    """
    src_db = get_src_db()
    ts = None
    found_meta = True

    if isinstance(col,DocMongoBackend):
        col = col.target_collection

    try:
        if col.database.name == config.DATA_TARGET_DATABASE:
            info = src_db["src_build"].find_one({"_id": col.name})
            if not info:
                logger.warning("Can't find information for target collection '%s'" % col.name)
            else:
                ts = info.get("_meta",{}).get("build_date")
                ts = ts and dtparser.parse(ts).timestamp()
        elif col.database.name == config.DATA_SRC_DATABASE:
            src_dump = get_src_dump()
            info = src_dump.find_one({"$where":"function() {if(this.upload) {for(var index in this.upload.jobs) {if(this.upload.jobs[index].step == \"%s\") return this;}}}" % col.name})
            if not info:
                logger.warning("Can't find information for source collection '%s'" % col.name)
            else:
                ts = info["upload"]["jobs"][col.name]["started_at"].timestamp()
        else:
            logging.warning("Can't find metadata for collection '%s' (not a target, not a source collection)" % col)
            found_meta = False
            build_cache = False
    except KeyError:
        logger.warning("Couldn't find timestamp in database for '%s'" % col.name)
    except Exception as e:
        logger.info("%s is not a mongo collection, _id cache won't be built (error: %s)" % (col,e))
        build_cache = False

    # try to find a cache file
    use_cache = False
    cache_file = None
    cache_format = getattr(config,"CACHE_FORMAT",None)
    if found_meta and getattr(config,"CACHE_FOLDER",None):
        cache_file = get_cache_filename(col.name)
        try:
            # size of empty file differs depending on compression
            empty_size = {None:0,"xz":32,"gzip":25,"bz2":14}
            if force_build:
                logger.warning("Force building cache file")
                use_cache = False
            # check size, delete if invalid
            elif os.path.getsize(cache_file) <= empty_size.get(cache_format,32): 
                logger.warning("Cache file exists but is empty, delete it")
                os.remove(cache_file)
            elif force_use:
                use_cache = True
                logger.info("Force using cache file")
            else:
                mt = os.path.getmtime(cache_file)
                if ts and mt >= ts:
                    dtmt = datetime.datetime.fromtimestamp(mt).isoformat()
                    dtts = datetime.datetime.fromtimestamp(ts).isoformat()
                    logging.debug("Cache is valid, modiftime_cache:%s >= col_timestamp:%s" % (dtmt,dtts))
                    use_cache = True
                else:
                    logger.info("Cache is too old, discard it")
        except FileNotFoundError:
            pass
    if use_cache:
        logger.debug("Found valid cache file for '%s': %s" % (col.name,cache_file))
        if validate_only:
            logging.debug("Only validating cache, now return")
            return []
        with open_compressed_file(cache_file) as cache_in:
            if cache_format:
                iocache = io.TextIOWrapper(cache_in)
            else:
                iocache = cache_in
            for ids in iter_n(iocache,batch_size):
                yield [_id.strip() for _id in ids if _id.strip()]
    else:
        logger.debug("No cache file found (or invalid) for '%s', use doc_feeder" % col.name)
        cache_out = None
        cache_temp = None
        if getattr(config,"CACHE_FOLDER",None) and config.CACHE_FOLDER and build_cache:
            if not os.path.exists(config.CACHE_FOLDER):
                os.makedirs(config.CACHE_FOLDER)
            cache_temp = "%s._tmp_" % cache_file
            # clean aborted cache file generation
            for tmpcache in glob.glob(os.path.join(config.CACHE_FOLDER,"%s*" % cache_temp)):
                logger.info("Removing aborted cache file '%s'" % tmpcache)
                os.remove(tmpcache)
            # use temp file and rename once done
            cache_temp = "%s%s" % (cache_temp,get_random_string())
            cache_out = get_compressed_outfile(cache_temp,compress=cache_format)
            logger.info("Building cache file '%s'" % cache_temp)
        else:
            logger.info("Can't build cache, cache not allowed or no cache folder")
            build_cache = False
        if isinstance(col,Collection):
            doc_feeder_func = partial(doc_feeder,col, step=batch_size, inbatch=True, fields={"_id":1})
        elif isinstance(col,DocMongoBackend):
            doc_feeder_func = partial(doc_feeder,col.target_collection, step=batch_size, inbatch=True, fields={"_id":1})
        elif isinstance(col,DocESBackend):
            # get_id_list directly return the _id, wrap it to match other 
            # doc_feeder_func returned vals. Also return a batch of id
            def wrap_id():
                ids = []
                for _id in col.get_id_list(step=batch_size):
                    ids.append({"_id":_id})
                    if len(ids) >= batch_size:
                        yield ids
                        ids = []
                if ids:
                    yield ids
            doc_feeder_func = partial(wrap_id)
        else:
            raise Exception("Unknown backend %s" % col)
        for doc_ids in doc_feeder_func():
            doc_ids = [str(_doc["_id"]) for _doc in doc_ids]
            if build_cache:
                strout = "\n".join(doc_ids) + "\n"
                if cache_format:
                    # assuming binary format (b/ccompressed)
                    cache_out.write(strout.encode())
                else:
                    cache_out.write(strout)
            yield doc_ids
        if build_cache:
            cache_out.close()
            cache_final = os.path.splitext(cache_temp)[0]
            try:
                os.rename(cache_temp,cache_final)
            except Exception as e:
                logger.exception("Couldn't set final cache filename, building cache failed")


def check_document_size(doc):
    """
    Return True if doc isn't too large for mongo DB
    """
    return len(bson.BSON.encode(doc)) < 16777216 #16*1024*1024

def get_previous_collection(new_id):
    """
    Given 'new_id', an _id from src_build, as the "new" collection,
    automatically select an "old" collection. By default, src_build's documents
    will be sorted according to their name (_id) and old colleciton is the one
    just before new_id.
    Note: because there can more than one build config used, the actual build config
    name is first determined using new_id collection name, then the find.sort is done
    on collections containing that build config name.
    """
    # TODO: this is not compatible with a generic hub_db backend
    # TODO: this should return a collection with status=success
    col = get_src_build()
    doc = col.find_one({"_id":new_id})
    assert doc, "No build document found for '%s'" % new_id
    assert "build_config" in doc, "No build configuration found for document '%s'" % new_id 
    assert doc["build_config"]["name"] == doc["build_config"]["_id"]
    confname = doc["build_config"]["name"]
    docs = get_src_build().find({
        "$and":[
            {"started_at":{"$lte":doc["started_at"]}},
            {"build_config.name":confname},
            {"archived":{"$exists":0}},
            ]},
        {"_id":1}).sort([("started_at",-1)]).limit(2)
    _ids = [d["_id"] for d in docs]
    assert len(_ids) == 2, "Expecting 2 collection _ids, got: %s" % _ids
    assert _ids[0] == new_id, "Can't find collection _id '%s'" % new_id
    return _ids[1]




"""
Elasticsearch Handlers

biothings.web.handlers.BaseESRequestHandler

    Supports: (all features above and)
    - access to biothing_type attribute
    - access to ES query pipeline stages
    - pretty print elasticsearch exceptions
    - common control option out_format

    Subclasses:
    - biothings.web.handlers.MetadataSourceHandler
    - biothings.web.handlers.MetadataFieldHandler
    - myvariant.web.beacon.BeaconHandler

biothings.web.handlers.ESRequestHandler

    Supports: (all features above and)
    - common control options (raw, rawquery)
    - common transform options (dotfield, always_list...)
    - query pipeline customization hooks
    - single query through GET
    - multiple quers through POST

    Subclasses:
    - biothings.web.handlers.BiothingHandler
    - biothings.web.handlers.QueryHandler

"""

from tornado.web import Finish

from biothings.utils.version import get_software_info
from biothings.utils.web.es import get_es_versions

from .api import BaseAPIHandler
from .exceptions import BadRequest, EndRequest

__all__ = [
    'BaseESRequestHandler',
    'MetadataSourceHandler',
    'MetadataFieldHandler',
    'ESRequestHandler',
    'BiothingHandler',
    'QueryHandler'
]


class BaseESRequestHandler(BaseAPIHandler):

    kwargs = {
        '*': {
            'out_format': {
                'type': str,
                'default': 'json',
                'enum': ('json', 'yaml', 'html', 'msgpack'),
                'alias': ['format']
            },
            'jsoninput': {
                'type': bool
            }  # deprecated
        }
    }

    def initialize(self, biothing_type=None):

        super(BaseESRequestHandler, self).initialize()
        self.biothing_type = biothing_type or self.web_settings.ES_DOC_TYPE
        self.query_builder = self.web_settings.query_builder
        self.query_backend = self.web_settings.query_backend
        self.result_transform = self.web_settings.result_transform

    def prepare(self):

        super().prepare()

        # supported across es requests
        self.format = self.args.out_format or 'json'

        # look up GA action tab in web settings
        # defined by request method and endpoint name
        action_key = f'GA_ACTION_{self.name.upper()}_{self.request.method}'
        if hasattr(self.web_settings, action_key):
            self.event['action'] = getattr(self.web_settings, action_key)

    def parse_exception(self, exception):

        message = super().parse_exception(exception)

        if '_es_error' in message:
            _es_error = message.pop('_es_error')
            message['error'] = _es_error.error
            try:
                root_cause = _es_error.info['error']['root_cause'][0]['reason']
                root_cause = root_cause.replace('\"', '\'').split('\n')
                for index, cause in enumerate(root_cause):
                    message['root_cuase_line_'+f'{index:02}'] = cause
            except Exception:
                self.logger.exception('Error parsing es exception.')

        return message


class MetadataSourceHandler(BaseESRequestHandler):
    """
    GET /metadata
    """
    name = 'metadata'
    kwargs = dict(BaseESRequestHandler.kwargs)
    kwargs['GET'] = {
        'dev': {'type': bool, 'default': False},
        'raw': {'type': bool, 'default': False}
    }

    async def get(self):

        _raw = await self.web_settings.read_index_mappings(self.biothing_type)
        _meta = self.web_settings.source_metadata[self.biothing_type]

        if self.args.raw:
            raise Finish(_raw)
        elif self.args.dev:
            _meta['software'] = get_software_info(
                app_dir=self.web_settings.get_git_repo_path())
            _meta['cluster'] = await get_es_versions(
                client=self.web_settings.async_es_client)
            _meta['hosts'] = self.web_settings.async_es_client.transport.hosts
        else:  # remove correlation info
            _meta.pop('_biothing', None)
            _meta.pop('_indices', None)

        _meta = self.extras(_meta)  # override here

        self.finish(dict(sorted(_meta.items())))

    def extras(self, _meta):
        """
        Override to add app specific metadata
        """
        return _meta


class MetadataFieldHandler(BaseESRequestHandler):
    """
    GET /metadata/fields
    """
    name = 'fields'
    kwargs = dict(BaseESRequestHandler.kwargs)
    kwargs['GET'] = {
        'raw': {'type': bool, 'default': False},
        'search': {'type': str, 'default': None},
        'prefix': {'type': str, 'default': None}
    }

    async def get(self):

        await self.web_settings.read_index_mappings(self.biothing_type)
        mapping = self.web_settings.source_properties[self.biothing_type]

        if self.args.raw:
            raise Finish(mapping)

        # reformat
        result = self.result_transform.transform_mapping(
            mapping, self.args.prefix, self.args.search)

        self.finish(result)


class ESRequestHandler(BaseESRequestHandler):
    '''
    Default Implementation of ES Query Pipelines
    '''
    # kwargs defined in biothing.web.settings.default
    kwarg_groups = ('control', 'esqb', 'es', 'transform')
    kwarg_methods = ('get', 'post')

    async def get(self, *args, **kwargs):
        return await self.execute_pipeline(*args, **kwargs)

    async def post(self, *args, **kwargs):
        return await self.execute_pipeline(*args, **kwargs)

    async def execute_pipeline(self, *args, **kwargs):

        options = self.pre_query_builder_hook(self.args)

        ###################################################
        #                   Build query
        ###################################################

        self._query = self.query_builder.build(options.esqb.q, options.esqb)
        self._query = self.pre_query_hook(options, self._query)

        ###################################################
        #                   Execute query
        ###################################################

        self._res = await self.query_backend.execute(self._query, options.es)
        self._res = self.pre_transform_hook(options, self._res)

        ###################################################
        #                 Transform result
        ###################################################

        res = self.result_transform.transform(self._res, options.transform)
        res = self.pre_finish_hook(options, res)

        self.finish(res)

    def pre_query_builder_hook(self, options):
        """
        Override this in subclasses.
        At this stage, we have the cleaned user input available.
        Might be a good place to implement input based tracking.
        """
        options.es.biothing_type = self.biothing_type
        options.transform.biothing_type = self.biothing_type

        # define multi-query response format
        if self.request.method == 'POST':
            queries = options.esqb.q
            options.transform.templates = (dict(query=q) for q in queries)
            options.transform.template_miss = dict(notfound=True)
            options.transform.template_hit = dict()

        return options

    def pre_query_hook(self, options, query):
        """
        Override this in subclasses.
        By default, return raw query, if requested.
        Might want to persist this behavior by calling super().
        """
        if options.control.rawquery:
            raise Finish(query.to_dict())
        return query

    def pre_transform_hook(self, options, res):
        """
        Override this in subclasses.
        By default, return query response, if requested.
        Might want to persist this behavior by calling super().
        """
        if options.control.raw:
            raise Finish(res)
        return res

    def pre_finish_hook(self, options, res):
        '''
        Override this in subclasses.
        Could implement additional result translation.
        '''
        return res


class BiothingHandler(ESRequestHandler):
    """
    Biothings Annotation Endpoint

    URL pattern examples:

        /{pre}/{ver}/{typ}/?
        /{pre}/{ver}/{typ}/([^\/]+)/?

        queries a term against a pre-determined field that
        represents the id of a document, like _id and dbsnp.rsid

        GET -> {...}
        POST -> [{...}, ...]
    """
    name = 'annotation'

    def pre_query_builder_hook(self, options):
        '''
        Annotation query has default scopes.
        Annotation query include _version field.
        '''
        if self.request.method == 'POST':
            self.event['qsize'] = len(options.ids)  # GA
            options.esqb.q = options.ids
        elif self.request.method == 'GET':
            options.esqb.q = options.id
        options.esqb.regexs = self.web_settings.ANNOTATION_ID_REGEX_LIST
        options.esqb.scopes = self.web_settings.ANNOTATION_DEFAULT_SCOPES
        options.esqb.version = True
        options = super().pre_query_builder_hook(options)
        return options

    def pre_finish_hook(self, options, res):
        """
        Return single result for GET.
        Empty results trigger 404 error.

        Keep _version field,
        Discard _score field.
        """
        if isinstance(res, dict):
            if not res.get('hits'):
                template = self.web_settings.ID_NOT_FOUND_TEMPLATE
                reason = template.format(bid=options.esqb.q)
                raise EndRequest(404, reason=reason)
            if len(res['hits']) > 1:
                raise EndRequest(404, reason='not a unique id.')
            res = res['hits'][0]
            res.pop('_score', None)

        elif isinstance(res, list):
            for hit in res:
                hit.pop('_score', None)

        return res


class QueryHandler(ESRequestHandler):
    '''
    Biothings Query Endpoint

    URL pattern examples:

        /{pre}/{ver}/{typ}/query/?
        /{pre}/{ver}//query/?

        GET -> {...}
        POST -> [{...}, ...]
    '''
    name = 'query'

    def pre_query_builder_hook(self, options):

        options = super().pre_query_builder_hook(options)

        # GA
        if self.request.method == 'GET':
            self.event['total'] = 0
        elif self.request.method == 'POST':
            self.event['qsize'] = len(options.esqb.q)

        # by default scroll ignores sort and size
        if options.es.fetch_all:
            options.esqb.pop('sort', None)
            options.esqb.pop('size', None)

        return options

    def pre_finish_hook(self, options, res):

        # GA
        if self.request.method == 'GET':
            if options.esqb.fetch_all:
                self.event['action'] = 'fetch_all'
                self.event['total'] = res.get('total', 0)

        return res



import types
import copy
import time
import logging

from pymongo.errors import DuplicateKeyError, BulkWriteError

from biothings.utils.common import timesofar, iter_n
from biothings.utils.dataload import merge_struct, merge_root_keys
from biothings.utils.mongo import get_src_db, check_document_size


class StorageException(Exception):
    pass


class BaseStorage(object):
    def __init__(self, db, dest_col_name, logger=logging):
        db = db or get_src_db()
        self.temp_collection = db[dest_col_name]
        self.logger = logger

    def process(self, iterable, *args, **kwargs):
        """
        Process iterable to store data. Must return the number
        of inserted records (even 0 if none)
        """
        raise NotImplementedError("implement-me in subclass")

    def check_doc_func(self, doc):
        """
        Return doc if it's alright, False if doc should be ignore for some reason
        Subclass and override as needed.
        """
        return doc


class CheckSizeStorage(BaseStorage):
    def check_doc_func(self, doc):
        ok = check_document_size(doc)
        # this is typically used to skip LFQSCWFLJHTTHZ-UHFFFAOYSA-N (Ethanol)
        # because there are too many elements in "ndc" list
        if not ok:
            self.logger.warning("Skip document '%s' because too large" %
                                doc.get("_id"))
            return False
        return ok


class BasicStorage(BaseStorage):
    def doc_iterator(self, doc_d, batch=True, batch_size=10000):
        if (isinstance(doc_d, types.GeneratorType)
                or isinstance(doc_d, list)) and batch:
            for doc_li in iter_n(doc_d, n=batch_size):
                doc_li = [d for d in doc_li if self.check_doc_func(d)]
                yield doc_li
        else:
            if batch:
                doc_li = []
                i = 0
            for _id, doc in doc_d.items():
                doc['_id'] = _id
                _doc = {}
                _doc.update(doc)
                if batch:
                    doc_li.append(_doc)
                    i += 1
                    if i % batch_size == 0:
                        doc_li = [d for d in doc_li if self.check_doc_func(d)]
                        yield doc_li
                        doc_li = []
                else:
                    yield self.check_doc_func(_doc)

            if batch:
                doc_li = [d for d in doc_li if self.check_doc_func(d)]
                yield doc_li

    def process(self, doc_d, batch_size):
        self.logger.info("Uploading to the DB...")
        t0 = time.time()
        total = 0
        for doc_li in self.doc_iterator(doc_d,
                                        batch=True,
                                        batch_size=batch_size):
            self.temp_collection.insert(doc_li,
                                        manipulate=False,
                                        check_keys=False)
            total += len(doc_li)
        self.logger.info('Done[%s]' % timesofar(t0))

        return total


class MergerStorage(BasicStorage):
    """
    This storage will try to merge documents when finding duplicated errors.
    It's useful when data is parsed using iterator. A record can be stored in database,
    then later, another record with the same ID is sent to the db, raising a duplicated error.
    These two documents would have been merged before using a 'put all in memory' parser.
    Since data is here read line by line, the merge is done while storing
    """

    merge_func = merge_struct

    def process(self, doc_d, batch_size):
        self.logger.info("Uploading to the DB...")
        t0 = time.time()
        tinner = time.time()
        aslistofdict = None
        total = 0
        for doc_li in self.doc_iterator(doc_d,
                                        batch=True,
                                        batch_size=batch_size):
            toinsert = len(doc_li)
            nbinsert = 0
            self.logger.info("Inserting %s records ... " % toinsert)
            try:
                bob = self.temp_collection.initialize_unordered_bulk_op()
                for d in doc_li:
                    aslistofdict = d.pop("__aslistofdict__", None)
                    bob.insert(d)
                res = bob.execute()
                nbinsert += res["nInserted"]
                self.logger.info("OK [%s]" % timesofar(tinner))
            except BulkWriteError as e:
                inserted = e.details["nInserted"]
                nbinsert += inserted
                self.logger.info("Fixing %d records " %
                                 len(e.details["writeErrors"]))
                ids = [d["op"]["_id"] for d in e.details["writeErrors"]]
                # build hash of existing docs
                docs = self.temp_collection.find({"_id": {"$in": ids}})
                hdocs = {}
                for doc in docs:
                    hdocs[doc["_id"]] = doc
                bob2 = self.temp_collection.initialize_unordered_bulk_op()
                for err in e.details["writeErrors"]:
                    errdoc = err["op"]
                    existing = hdocs[errdoc["_id"]]
                    if errdoc is existing:
                        # if the same document has been yielded twice,
                        # they could be the same, so we ignore it but
                        # count it as processed (see assert below)
                        nbinsert += 1
                        continue
                    assert "_id" in existing
                    _id = errdoc.pop("_id")
                    merged = self.__class__.merge_func(
                        errdoc, existing, aslistofdict=aslistofdict)
                    # update previously fetched doc. if several errors are about the same doc id,
                    # we would't merged things properly without an updated document
                    assert "_id" in merged
                    bob2.find({"_id": _id}).update_one({"$set": merged})
                    hdocs[_id] = merged
                    nbinsert += 1

                res = bob2.execute()
                self.logger.info("OK [%s]" % timesofar(tinner))
            assert nbinsert == toinsert, "nb %s to %s" % (nbinsert, toinsert)
            # end of loop so it counts the time spent in doc_iterator
            tinner = time.time()
            total += nbinsert

        self.logger.info('Done[%s]' % timesofar(t0))

        return total


class RootKeyMergerStorage(MergerStorage):
    """
    Just like MergerStorage, this storage deals with duplicated error
    by appending key's content to existing document. Keys in existing
    document will be converted to a list as needed.

    Note:
      - root keys must have the same type in each documents
      - inner structures aren't merged together, the merge happend
        at root key level
    """
    @classmethod
    def merge_func(klass, doc1, doc2, **kwargs):
        # caller popped it from doc1, let's take from doc2
        _id = doc2["_id"]
        # exclude_id will remove _if from doc2, that's why we kept it from before
        # also, copy doc2 ref as the merged doc will be stored in
        # a bulk op object, since doc2 is modified in place, this could impact
        # the bulk op and procude empty $set error from mongo
        doc = merge_root_keys(doc1, copy.copy(doc2), exclude=["_id"])
        doc["_id"] = _id
        return doc


class IgnoreDuplicatedStorage(BasicStorage):
    def process(self, iterable, batch_size):
        self.logger.info("Uploading to the DB...")
        t0 = time.time()
        tinner = time.time()
        total = 0
        for doc_li in self.doc_iterator(iterable,
                                        batch=True,
                                        batch_size=batch_size):
            try:
                bob = self.temp_collection.initialize_unordered_bulk_op()
                for d in doc_li:
                    bob.insert(d)
                res = bob.execute()
                total += res['nInserted']
                self.logger.info("Inserted %s records [%s]" %
                                 (res['nInserted'], timesofar(tinner)))
            except BulkWriteError as e:
                self.logger.info(
                    "Inserted %s records, ignoring %d [%s]" %
                    (e.details['nInserted'], len(
                        e.details["writeErrors"]), timesofar(tinner)))
            except Exception:
                raise
            tinner = time.time()
        self.logger.info('Done[%s]' % timesofar(t0))

        return total


class NoBatchIgnoreDuplicatedStorage(BasicStorage):
    """
    You should use IgnoreDuplicatedStorag, which works using batch
    and is thus way faster...
    """
    def process(self, doc_d, batch_size):
        self.logger.info("Uploading to the DB...")
        t0 = time.time()
        tinner = time.time()
        # force step = 1
        cnt = 0
        total = 0
        dups = 0
        for doc_li in self.doc_iterator(doc_d, batch=True, batch_size=1):
            try:
                self.temp_collection.insert(doc_li,
                                            manipulate=False,
                                            check_keys=False)
                cnt += 1
                total += 1
                if (cnt + dups) % batch_size == 0:
                    # we insert one by one but display progress on a "batch_size" base
                    self.logger.info("Inserted %s records, ignoring %s [%s]" %
                                     (cnt, dups, timesofar(tinner)))
                    cnt = 0
                    dups = 0
                    tinner = time.time()
            except DuplicateKeyError:
                dups += 1
                pass
        self.logger.info('Done[%s]' % timesofar(t0))

        return total


class UpsertStorage(BasicStorage):
    """Insert or update documents, based on _id"""
    def process(self, iterable, batch_size):
        self.logger.info("Uploading to the DB...")
        t0 = time.time()
        tinner = time.time()
        total = 0
        for doc_li in self.doc_iterator(iterable,
                                        batch=True,
                                        batch_size=batch_size):
            try:
                bob = self.temp_collection.initialize_unordered_bulk_op()
                for d in doc_li:
                    bob.find({"_id": d["_id"]}).upsert().replace_one(d)
                res = bob.execute()
                nb = res["nUpserted"] + res["nModified"]
                total += nb
                self.logger.info("Upserted %s records [%s]" %
                                 (nb, timesofar(tinner)))
            except Exception:
                raise
            tinner = time.time()
        self.logger.info('Done[%s]' % timesofar(t0))

        return total


class NoStorage(object):
    """
    This a kind of a place-holder, this storage will just store nothing...
    (but it will respect storage interface)
    """
    def __init__(self, db_info, dest_col_name, logger):
        db = get_src_db()
        self.temp_collection = db[dest_col_name]
        self.logger = logger

    def process(self, iterable, *args, **kwargs):
        self.logger.info("NoStorage stores nothing, skip...")
        return 0



# from http://asyncssh.readthedocs.io/en/latest/#id13

# To run this program, the file ``ssh_host_key`` must exist with an SSH
# private key in it to use as a server host key.

import sys
import os
import re
import datetime
import json
import pydoc
import copy
import asyncio
import io
import time
from functools import partial
from pprint import pformat
from collections import OrderedDict

from IPython import InteractiveShell

from biothings.utils.dataload import to_boolean
from biothings.utils.common import timesofar
import biothings.utils.aws as aws
from biothings.utils.hub_db import get_cmd, get_last_command, backup, restore
from biothings.utils.loggers import ShellLogger
from biothings.utils.web.es import flatten_doc
from biothings import config
if hasattr(config, 'logger'):
    logging = config.logger
else:
    import logging

# useful variables to bring into hub namespace
pending = "pending"

HUB_ENV = hasattr(config, "HUB_ENV") and config.HUB_ENV or ""    # default: prod (or "normal")
VERSIONS = HUB_ENV and "%s-versions" % HUB_ENV or "versions"
LATEST = HUB_ENV and "%s-latest" % HUB_ENV or "latest"


def jsonreadify(cmd):
    newcmd = copy.copy(cmd)
    newcmd.pop("jobs")
    return newcmd

##############
# HUB SERVER #
##############

class AlreadyRunningException(Exception):
    pass

class CommandError(Exception):
    pass

class NoSuchCommand(Exception):
    pass

class CommandNotAllowed(Exception):
    pass

class CommandInformation(dict):
    pass

class CommandDefinition(dict):
    pass


class HubShell(InteractiveShell):

    launched_commands = {}
    pending_outputs = {}
    cmd_cnt = None
    cmd = None  # "cmd" collection

    def __init__(self, job_manager):
        self.job_manager = job_manager
        self.commands = OrderedDict()
        self.managers = {}
        self.extra_ns = OrderedDict()
        self.tracked = {}   # command calls kept in history or not
        self.hidden = {}    # not returned by help()
        self.origout = sys.stdout
        self.buf = io.StringIO()
        self.shellog = ShellLogger(name="shell")
        # there should be only one shell instance (todo: singleton)
        self.__class__.cmd = get_cmd()
        self.__class__.set_command_counter()
        super(HubShell, self).__init__(user_ns=self.extra_ns)

    @classmethod
    def set_command_counter(cls):
        assert cls.cmd, "No cmd collection set"
        try:
            res = get_last_command()
            if res:
                logging.debug("Last launched command ID: %s", res["_id"])
                cls.cmd_cnt = int(res["_id"]) + 1
            else:
                logging.info("No previously stored command found, set counter to 1")
                cls.cmd_cnt = 1

        except StopIteration:
            logging.info("Can't find highest command number, assuming starting from scratch")
            cls.cmd_cnt = 1

    def set_commands(self, basic_commands, *extra_ns):
        def register(commands, hidden=False):
            for name, cmd in commands.items():
                if name in self.commands:
                    raise CommandError("Command defined multiple time: %s" % name)
                # if type(cmd) == CommandDefinition:      # TODO: remove this line
                if isinstance(cmd, CommandDefinition):
                    try:
                        self.commands[name] = cmd["command"]
                        self.tracked[name] = cmd.get("tracked", True)
                        self.hidden[name] = cmd.get("hidden", False) or hidden
                    except KeyError as e:
                        raise CommandError("Could not register command because missing '%s' in definition: %s" % (e, cmd))
                else:
                    self.commands[name] = cmd
                    self.hidden[name] = hidden
                # update original passed commands to caller knows what's been done there
                if name not in basic_commands:
                    basic_commands[name] = cmd
        # update with ssh server default commands
        register(basic_commands)
        # don't track this calls
        register({"restart": CommandDefinition(command=self.restart, track=True)})
        register({"stop": CommandDefinition(command=self.stop, track=True)})
        register({"backup": CommandDefinition(command=backup, track=True)})
        register({"restore": CommandDefinition(command=restore, track=True)})
        register({"help": CommandDefinition(command=self.help, track=False)})
        register({"commands": CommandDefinition(command=self.command_info, tracked=False)})
        register({"command": CommandDefinition(
            command=lambda id, *args, **kwargs: self.command_info(id=id, *args, **kwargs),
            tracked=False
        )})

        for extra in extra_ns:
            # don't expose extra commands, they're kind of private/advanced
            register(extra, hidden=True)

        #self.extra_ns["cancel"] = self.__class__.cancel
        # for boolean calls
        self.extra_ns["_and"] = _and
        self.extra_ns["partial"] = partial
        self.extra_ns["hub"] = self
        # merge official/public commands with hidden/private to
        # make the whole available in shell's namespace
        self.extra_ns.update(self.commands)
        # Note: there's no need to update shell namespace as self.extra_ns
        # has been passed by ref in __init__() so things get updated automagically
        # (self.user_ns.update(...) can be used otherwise, self.user_ns is IPython
        # internal namespace dict

    def stop(self, force=False):
        return self.restart(force=force, stop=True)

    def restart(self, force=False, stop=False):

        @asyncio.coroutine
        def do():
            try:
                if stop:
                    event = "hub_stop"
                    msg = "Hub is stopping"
                else:
                    event = "hub_restart"
                    msg = "Hub is restarting"
                logging.critical(
                    json.dumps({
                        "type": "alert",
                        "msg": msg,
                        "event": event
                    }),
                    extra={"event": True}
                )
                logging.info("Stopping job manager...")
                j = self.job_manager.stop(force=force)
                def ok(f):
                    f.result()  # consume
                    logging.error("Job manager stopped")
                j.add_done_callback(ok)
                yield from j
            except Exception as e:
                logging.error("Error while recycling the process queue: %s", e)
                raise

        def start(f):
            f.result()  # consume future's result to potentially raise exception
            logging.debug("%s %s", [sys.executable], sys.argv)
            import subprocess
            subprocess.Popen([sys.executable] + sys.argv)
            self.job_manager.hub_process.kill()
            sys.exit(0)

        def autokill(f):
            f.result()
            self.job_manager.hub_process.kill()

        fut = asyncio.ensure_future(do())

        if stop:
            logging.warning("Stopping hub")
            fut.add_done_callback(autokill)
        else:
            logging.warning("Restarting hub")
            fut.add_done_callback(start)

        return fut

    def help(self, func=None):
        """
        Display help on given function/object or list all available commands
        """
        if not func:
            cmds = "\nAvailable commands:\n\n"
            for k in self.commands:
                if self.hidden[k]:
                    continue
                cmds += "\t%s\n" % k
            cmds += "\nType: 'help(command)' for more\n"
            return cmds
        elif isinstance(func, partial):
            docstr = "\n" + pydoc.render_doc(func.func, title="Hub documentation: %s")
            docstr += "\nDefined et as a partial, with:\nargs:%s\nkwargs:%s\n" % (repr(func.args), repr(func.keywords))
            return docstr
        elif isinstance(func, CompositeCommand):
            docstr = "\nComposite command:\n\n%s\n" % func
            return docstr
        else:
            try:
                return "\n" + pydoc.render_doc(func, title="Hub documentation: %s")
            except ImportError:
                return "\nHelp not available for this command\n"

    def launch(self, pfunc):
        """
        Helper to run a command and register it
        pfunc is partial taking no argument. Command name
        is generated from partial's func and arguments
        """
        res = pfunc()
        # rebuild a command as string
        strcmd = pfunc.func.__name__
        strcmd += "("
        strargs = []
        if pfunc.args:
            strargs.append(",".join([repr(a) for a in pfunc.args]))
        if pfunc.keywords:
            strargs.append(",".join(["%s=%s" % (k, repr(v)) for (k, v) in pfunc.keywords.items()]))
        strcmd += ",".join(strargs)
        strcmd += ")"
        # we use force here because, very likely, the command from strcmd we generated
        # isn't part of shell's known commands (and there's a check about that when force=False)
        self.register_command(strcmd, res, force=True)
        return pfunc

    def extract_command_name(self, cmd):
        try:
            # extract before () (non-callable are never tracked)
            grps = re.fullmatch("([\w\.]+)(\(.*\))", cmd.strip()).groups()
            return grps[0]
        except AttributeError:
            raise CommandError("Can't extract command name from '%s'" % repr(cmd))

    @classmethod
    def save_cmd(cls, _id, cmd):
        newcmd = jsonreadify(cmd)
        cls.cmd.replace_one({"_id": _id}, newcmd, upsert=True)

    def register_managers(self, managers):
        self.managers = managers

    def register_command(self, cmd, result, force=False):
        """
        Register a command 'cmd' inside the shell (so we can keep track of it).
        'result' is the original value that was returned when cmd was submitted.
        Depending on the type, returns a cmd number (ie. result was an asyncio task
        and we need to wait before getting the result) or directly the result of
        'cmd' execution, returning, in that case, the output.
        """
        # see if command should actually be registered
        try:
            cmdname = self.extract_command_name(cmd)
        except CommandError:
            # if can't extract command name, then don't even try to register
            # (could be, for instance, "pure" python code typed from the console)
            logging.debug("Can't extract command from %s, can't register", cmd)
            return result
        # also, never register non-callable command
        if not force and (not callable(self.extra_ns.get(cmdname)) or \
                          self.tracked.get(cmdname, True) == False):
            return result

        cmdnum = self.__class__.cmd_cnt
        cmdinfo = CommandInformation(cmd=cmd, jobs=result, started_at=time.time(),
                                     id=cmdnum, is_done=False)
        assert cmdnum not in self.__class__.launched_commands
        # register
        self.__class__.launched_commands[cmdnum] = cmdinfo
        self.__class__.save_cmd(cmdnum, cmdinfo)
        self.__class__.cmd_cnt += 1

        # TODO: clean up the following if statement, confirm with Sebastien
        if type(result) == asyncio.tasks.Task or type(result) == asyncio.tasks._GatheringFuture or \
                type(result) == asyncio.Future or \
                type(result) == list and len(result) > 0 and type(result[0]) == asyncio.tasks.Task:
            # it's asyncio related
            result = type(result) != list and [result] or result   # TODO: cleanup and confirm this line
            cmdinfo["jobs"] = result
            return cmdinfo
        else:
            # ... and it's not asyncio related, we can display it directly
            cmdinfo["is_done"] = True
            cmdinfo["failed"] = False
            cmdinfo["started_at"] = time.time()
            cmdinfo["finished_at"] = time.time()
            cmdinfo["duration"] = "0s"
            return result

    def eval(self, line, return_cmdinfo=False, secure=False):
        line = line.strip()
        self.shellog.input(line)
        origline = line  # keep what's been originally entered
        # poor man's singleton...
        if line in [j["cmd"] for j in self.__class__.launched_commands.values() if not j.get("is_done")]:
            raise AlreadyRunningException("Command '%s' is already running\n" % repr(line))
        # is it a hub command, in which case, intercept and run the actual declared cmd
        # IMPORTANT !!! this is where we allow the command or not when secure=True IMPORTANT !!!
        # the logic is following:
        # - what's before parenthesis must exactly match a command
        # - parenthesis are mandatory
        # - no '&&' operator allowed
        if secure:
            # command must be alpha only, argument with "," and "=", or no arg at all
            pat = r'^([A-Za-z_]+)\(["\'\w\s=,.-]*\)$'
        else:
            pat = r'(.*)\(.*\)'  # more permissive
        m = re.match(pat, line)
        if m:
            cmd = m.groups()[0].strip()
            if secure and cmd not in self.commands:
                #  match regex rule but not a valid/existing command, discard it
                raise NoSuchCommand(cmd)
            if cmd in self.commands and \
                    isinstance(self.commands[cmd], CompositeCommand):
                line = self.commands[cmd]
        elif line != "" and secure:
            # we have something entered, it doesn't match our regex rule, discard it
            raise CommandNotAllowed(line)
        # cmdline is the actual command sent to shell, line is the one displayed
        # they can be different if there's a preprocessing
        cmdline = line
        # && cmds ? ie. chained cmds
        if "&&" in line:
            chained_cmds = [cmd for cmd in map(str.strip, line.split("&&")) if cmd]
            if len(chained_cmds) > 1:
                # need to build a command with _and and using partial, meaning passing original func param
                # to the partials
                strcmds = []
                for one_cmd in chained_cmds:
                    func, args = re.match("(.*)\((.*)\)", one_cmd).groups()
                    if args:
                        strcmds.append("partial(%s,%s)" % (func, args))
                    else:
                        strcmds.append("partial(%s)" % func)
                cmdline = "_and(%s)" % ",".join(strcmds)
            else:
                raise CommandError("Using '&&' operator required two operands\n")
        r = self.run_cell(cmdline, store_history=True)
        outputs = []
        if not r.success:
            raise CommandError("%s\n" % repr(r.error_in_exec))
        else:
            # command was a success, now get the results:
            if r.result is None:
                # -> nothing special was returned, grab the stdout
                self.buf.seek(0)
                # from print stdout ?
                b = self.buf.read()
                outputs.append(b)
                # clear buffer
                self.buf.seek(0)
                self.buf.truncate()
            else:
                # -> we have something returned...
                res = self.register_command(cmd=origline, result=r.result)
                # if type(res) != CommandInformation:    # TODO: remove this line
                if not isinstance(res, CommandInformation):
                    # if type(res) != str:    # TODO: remove this line
                    if not isinstance(res, str):
                        outputs.append(pformat(res))
                    else:
                        outputs.append(res)
                else:
                    if return_cmdinfo:
                        return res

        # Note: this will cause all outputs to go to one SSH session, ie. if multiple users
        # are logged, only one will the results
        if self.__class__.pending_outputs:
            outputs.extend(self.__class__.pending_outputs.values())
            self.__class__.pending_outputs = {}

        return outputs

    #@classmethod
    #def cancel(klass,jobnum):
    #    return klass.launched_commands.get(jobnum)

    @classmethod
    def refresh_commands(cls):
        for num, info in sorted(cls.launched_commands.items()):
            # already process, this current command is now history
            # Note: if we have millions of commands there, it could last quite a while,
            # but IRL we only have a few
            if info.get("is_done") == True:
                continue
            # is_done = set([j.done() for j in info["jobs"]]) == set([True])   # TODO: remove this line
            is_done = {j.done() for j in info["jobs"]} == {True}
            has_err = is_done and [True for j in info["jobs"] if j.exception()] or None
            localoutputs = is_done and ([str(j.exception()) for j in info["jobs"] if j.exception()] or \
                        [j.result() for j in info["jobs"]]) or None
            if is_done:
                cls.launched_commands[num]["is_done"] = True
                cls.launched_commands[num]["failed"] = has_err and has_err[0] or False
                cls.launched_commands[num]["results"] = localoutputs
                cls.launched_commands[num]["finished_at"] = time.time()
                cls.launched_commands[num]["duration"] = timesofar(t0=cls.launched_commands[num]["started_at"],
                                                                   t1=cls.launched_commands[num]["finished_at"])
                cls.save_cmd(num, cls.launched_commands[num])
                if not has_err and localoutputs and set(map(type, localoutputs)) == {str}:
                    localoutputs = "\n" + "".join(localoutputs)
                cls.pending_outputs[num] = "[%s] %s {%s} %s: finished %s " % \
                        (num, has_err and "ERR" or "OK", timesofar(info["started_at"]), info["cmd"], localoutputs)
            else:
                cls.pending_outputs[num] = "[%s] RUN {%s} %s" % (num, timesofar(info["started_at"]), info["cmd"])

    @classmethod
    def command_info(cls, id=None, running=None, failed=None):
        cmds = {}
        if id is not None:
            try:
                id = int(id)
                return jsonreadify(cls.launched_commands[id])
            except KeyError:
                raise CommandError("No such command with ID %s" % repr(id))
            except ValueError:
                raise CommandError("Invalid ID %s" % repr(id))
        if running is not None:
            is_done = not to_boolean(running)
        else:
            is_done = None
        if failed is not None:
            failed = to_boolean(failed)
        for _id, cmd in cls.launched_commands.items():
            if is_done is not None:
                # running or done commands (not both)
                if cmd.get("is_done") == is_done:
                    # done + failed (a failed command is always done btw)
                    if failed is not None and cmd.get("is_done") == True:
                        if cmd.get("failed") == failed:
                            cmds[_id] = jsonreadify(cmd)
                    else:
                        # don't care if failed or not
                        cmds[_id] = jsonreadify(cmd)
                else:
                    # If asked is_done=true, it means command _id has is_done=false
                    # if we get there. So the command is sill running, so we don't
                    # know if it failed or not, so no need to check failed there,
                    # it's been handled above.
                    # If asksed is_done=false, we don't need to check failed,
                    # same logic applies
                    continue
            else:
                # either running or done commands (both)
                if failed is not None and cmd.get("is_done") == True:
                    if cmd.get("failed") == failed:
                        cmds[_id] = jsonreadify(cmd)
                else:
                    # don't care if failed or not
                    cmds[_id] = jsonreadify(cmd)

        return cmds


####################
# DEFAULT HUB CMDS #
####################
# these can be used in client code to define
# commands. partial should be used to pass the
# required arguments, eg.:
# {"schedule" ; partial(schedule,loop)}

def stats(src_dump):
    pass


def template_out(field, confdict):
    """
    Return field as a templated-out filed,
    substituting some "%(...)s" part with confdict,
    Fields can follow dotfield notation.
    Fields like "$(...)" are replaced with a timestamp
    following specified format (see time.strftime)
    Example::

        confdict = {"a":"one"}
        field = "%(a)s_two_three_$(%Y%m)"
        => "one_two_three_201908" # assuming we're in August 2019
    """
    # first deal with timestamp
    pat = re.compile(".*(\$\((.*?)\)).*")

    try:
        m = pat.match(field)
    except TypeError:
        # not string/byte-like just skip the process
        return field

    if m:
        tosub, fmt = m.groups()
        ts = datetime.datetime.now().strftime("%%%s" % fmt)
        field = field.replace(tosub, ts)
    flatdict = flatten_doc(confdict)
    # then use dict to sub keys
    field = field % flatdict

    return field


def publish_data_version(s3_bucket, s3_folder, version_info, update_latest=True,
                         aws_key=None, aws_secret=None):
    """
    Update remote files:
        - versions.json: add version_info to the JSON list
                        or replace if arg version_info is a list
        - latest.json: update redirect so it points to latest version url

    "versions" is dict such as::

        {"build_version":"...",         # version name for this release/build
         "require_version":"...",       # version required for incremental update
         "target_version": "...",       # version reached once update is applied
         "type" : "incremental|full"    # release type
         "release_date" : "...",        # ISO 8601 timestamp, release date/time
         "url": "http...."}             # url pointing to release metadata
    """
    # register version
    versionskey = os.path.join(s3_folder, "%s.json" % VERSIONS)
    try:
        versions = aws.get_s3_file(versionskey,
                                   return_what="content",
                                   aws_key=aws_key,
                                   aws_secret=aws_secret,
                                   s3_bucket=s3_bucket)
        versions = json.loads(versions.decode())  # S3 returns bytes
    except (FileNotFoundError, json.JSONDecodeError):
        versions = {"format": "1.0", "versions": []}
    # if type(version_info) == list:     # remove this line
    if isinstance(version_info, list):
        versions["versions"] = version_info
    else:
        # used to check duplicates
        tmp = {}
        # [tmp.setdefault(e["build_version"], e) for e in versions["versions"]]   # TODO: remove this line
        for e in versions["versions"]:
            tmp.setdefault(e["build_version"], e)
        tmp[version_info["build_version"]] = version_info
        # order by build_version
        versions["versions"] = sorted(tmp.values(), key=lambda e: e["build_version"])

    aws.send_s3_file(None, versionskey,
                     content=json.dumps(versions, indent=True),
                     aws_key=aws_key,
                     aws_secret=aws_secret,
                     s3_bucket=s3_bucket,
                     content_type="application/json",
                     overwrite=True)

    # update latest
    # if type(version_info) != list and update_latest:    # TODO: remove this line
    if not isinstance(version_info, list) and update_latest:
        latestkey = os.path.join(s3_folder, "%s.json" % LATEST)
        key = None
        try:
            key = aws.get_s3_file(
                latestkey,
                return_what="key",
                aws_key=aws_key,
                aws_secret=aws_secret,
                s3_bucket=s3_bucket
            )
        except FileNotFoundError:
            pass
        aws.send_s3_file(
            None, latestkey,
            content=json.dumps(version_info["build_version"], indent=True),
            content_type="application/json",
            aws_key=aws_key,
            aws_secret=aws_secret,
            s3_bucket=s3_bucket,
            overwrite=True
        )
        if not key:
            key = aws.get_s3_file(
                latestkey,
                return_what="key",
                aws_key=aws_key,
                aws_secret=aws_secret,
                s3_bucket=s3_bucket
            )
        newredir = os.path.join("/", s3_folder, "{}.json".format(version_info["build_version"]))
        key.set_redirect(newredir)


def _and(*funcs):
    """
    Calls passed functions, one by one. If one fails, then it stops.
    Function should return a asyncio Task. List of one Task only are also permitted.
    Partial can be used to pass arguments to functions.
    Ex: _and(f1,f2,partial(f3,arg1,kw=arg2))
    """
    all_res = []
    func1 = funcs[0]
    func2 = None
    fut1 = func1()
    # if type(fut1) == list:   # TODO: remove this line
    if isinstance(fut1, list):
        assert len(fut1) == 1, "Can't deal with list of more than 1 task: %s" % fut1
        fut1 = fut1.pop()
    if not isinstance(fut1, asyncio.Future):
        raise CommandError("First command didn't return a future, can't chain commands")
    all_res.append(fut1)
    # err = None
    def do(f, cb):
        f.result()  # consume exception if any
        if cb:
            all_res.extend(_and(cb, *funcs))
    if len(funcs) > 1:
        func2 = funcs[1]
        if len(funcs) > 2:
            funcs = funcs[2:]
        else:
            funcs = []
    fut1.add_done_callback(partial(do, cb=func2))
    return all_res


class CompositeCommand(str):
    """
    Defines a composite hub commands, that is,
    a new command made of other commands. Useful to define
    shortcuts when typing commands in hub console.
    """
    def __init__(self, cmd):
        self.cmd = cmd

    def __str__(self):
        return "<CompositeCommand: '%s'>" % self.cmd

############
# RELOADER #
############

def exclude_from_reloader(path):
    # exlucde cached, git and hidden files
    return path.endswith("__pycache__") or \
           ".git" in path or \
           os.path.basename(path).startswith(".")


class BaseHubReloader(object):
    """
    Monitor sources' code and reload hub accordingly to update running code
    """

    def __init__(self, paths, reload_func, wait=5.0):
        """
        Monitor given paths for directory deletion/creation
        and for file deletion/creation. Poll for events every 'wait' seconds.
        """
        raise NotImplementedError("Implement me in a sub-class")

    def poll(self):
        """Start monitoring changes on files and/directories"""
        raise NotImplementedError("Implement me in a sub-class")

    def watched_files(self):
        """Return a list of files/directories being watched"""
        raise NotImplementedError("Implement me in a sub-class")


class PyInotifyHubReloader(BaseHubReloader):
    """Based on pyinotify events"""

    # inner class to hide pyinotify in case not available
    import pyinotify
    class ReloadListener(pyinotify.ProcessEvent):
    
        def my_init(self, watcher_manager, reload_func):
            self.reload_func = reload_func
            self.watcher_manager = watcher_manager
    
        def process_default(self, event):
            pyinotify = sys.modules["pyinotify"]
            if exclude_from_reloader(event.pathname):
                return
            if event.dir:
                if event.mask & pyinotify.IN_CREATE:
                    # add to watcher. no need to check if already watched, manager knows
                    # how to deal with that
                    logging.info("Add '%s' to watcher", event.pathname)
                    self.watcher_manager.add_watch(event.pathname, self.notifier.mask, rec=1)
                elif event.mask & pyinotify.IN_DELETE:
                    logging.info("Remove '%s' from watcher", event.pathname)
                    # watcher knows when directory is deleted (file descriptor become invalid),
                    # so no need to do it manually
            logging.error("Need to reload manager because of %s", event)
            self.reload_func()
    

    def __init__(self, paths, reload_func, wait=5, mask=None):
        pyinotify = sys.modules["pyinotify"] # get it from sys.modules or we'd need another "import"
                                             # just sure why...
        if isinstance(paths, str):
            paths = [paths]
        paths = set(paths)  # get rid of duplicated, just in case
        self.mask = mask or pyinotify.IN_CREATE|pyinotify.IN_DELETE|pyinotify.IN_CLOSE_WRITE
        self.reload_func = reload_func
        # only listen to these events. Note: directory detection is done via a flag so
        # no need to use IS_DIR
        self.watcher_manager = pyinotify.WatchManager(exclude_filter=exclude_from_reloader)
        self.paths = []  # cleaned
        for path in paths:
            if not os.path.isabs(path):
                path = os.path.abspath(path)
            self.paths.append(path)
            self.watcher_manager.add_watch(path, self.mask, rec=True, exclude_filter=exclude_from_reloader)   # recursive

        self.listener = self.__class__.ReloadListener(
                watcher_manager=self.watcher_manager,
                reload_func=self.reload_func)
        self.notifier = pyinotify.Notifier(
                self.watcher_manager,
                default_proc_fun=self.listener
                )
        # propagate notifier so notifer itself can be reloaded (when new directory/source)
        self.listener.notifier = self
        self.wait = wait
        self.do_monitor = False

    def monitor(self):

        @asyncio.coroutine
        def do():
            logging.info(
                "Monitoring source code in, %s:\n%s",
                repr(self.paths),
                pformat(self.watched_files())
            )
            while self.do_monitor:
                try:
                    yield from asyncio.sleep(self.wait)
                    # this reads events from OS
                    if self.notifier.check_events(0.1):
                        self.notifier.read_events()
                    # Listener gets called there if event avail
                    self.notifier.process_events()
                except KeyboardInterrupt:
                    logging.warning("Stop monitoring code")
                    break

        self.do_monitor = True
        return asyncio.ensure_future(do())

    def watched_files(self):
        return [v.path for v in self.watcher_manager.watches.values()]


class TornadoAutoReloadHubReloader(BaseHubReloader):
    """Reloader based on tornado.autoreload module"""

    def __init__(self, paths, reload_func, wait=5):
        self.mod = sys.modules["tornado.autoreload"]
        if isinstance(paths, str):
            paths = [paths]
        paths = set(paths)  # get rid of duplicated, just in case
        self.reload_func = reload_func
        self.mod.add_reload_hook(self.reload_func)
        # only listen to these events. Note: directory detection is done via a flag so
        # no need to use IS_DIR
        self.paths = []  # cleaned
        for path in paths:
            if not os.path.isabs(path):
                path = os.path.abspath(path)
            self.paths.append(path)
            self.mod.watch(path)
            for dirpath, dirnames, filenames in os.walk(path):
                if exclude_from_reloader(dirpath):
                    continue
                if not dirnames:
                    # reaching files (leaves)
                    for fn in filenames:
                        f_path = os.path.join(dirpath, fn)
                        if exclude_from_reloader(f_path):
                            continue
                        self.mod.watch(f_path)
        self.wait = wait
        self.do_monitor = False

    def monitor(self):
        logging.info(
            "Monitoring source code in, %s:\n%s",
            repr(self.paths),
            pformat(self.watched_files())
        )
        self.mod.start(self.wait*1000) # millis

    def watched_files(self):
        return self.mod._watched_files
        #return [d for d in self.mod._watched_files if os.path.isdir(d)]



def get_hub_reloader(*args, **kwargs):
    """
    Select proper reloader depending on whether pyinotify is
    available (1st choice) or not (tornado.autoreload otherwise)
    """
    if getattr(config, "USE_RELOADER", False) and config.USE_RELOADER:
        try:
            import pyinotify
            logging.info("Using Hub reloader based on pyinotify")
            return PyInotifyHubReloader(*args,**kwargs)
        except ImportError:
            import tornado.autoreload
            logging.info("Using Hub reloader based on tornado.autoreload")
            return TornadoAutoReloadHubReloader(*args,**kwargs)
    else:
        logging.info("USE_RELOADER not set (or False), won't monitor for changes")
        return None



from dateutil.parser import parse as dtparse
import locale

locale.setlocale(locale.LC_ALL, '')


class ReleaseNoteTxt(object):
    def __init__(self, changes):
        self.changes = changes
        #pprint(self.changes)

    def save(self, filepath):
        try:
            import prettytable
        except ImportError:
            raise ImportError(
                "Please install prettytable to use this rendered")

        def format_number(n, sign=None):
            s = ""
            if sign:
                if n > 0:
                    s = "+"
                elif n < 0:
                    s = "-"
            try:
                n = abs(n)
                strn = "%s%s" % (s, locale.format("%d", n, grouping=True))
            except TypeError:
                # something wrong with converting, maybe we don't even have a number to format...
                strn = "N.A"
            return strn

        txt = ""
        title = "Build version: '%s'" % self.changes["new"]["_version"]
        txt += title + "\n"
        txt += "".join(["="] * len(title)) + "\n"
        dt = dtparse(self.changes["generated_on"])
        txt += "Previous build version: '%s'\n" % self.changes["old"][
            "_version"]
        txt += "Generated on: %s\n" % dt.strftime("%Y-%m-%d at %H:%M:%S")
        txt += "\n"

        table = prettytable.PrettyTable([
            "Updated datasource", "prev. release", "new release",
            "prev. # of docs", "new # of docs"
        ])
        table.align["Updated datasource"] = "l"
        table.align["prev. release"] = "c"
        table.align["new release"] = "c"
        table.align["prev. # of docs"] = "r"
        table.align["new # of docs"] = "r"

        for src, info in sorted(self.changes["sources"]["added"].items(),
                                key=lambda e: e[0]):
            main_info = dict([(k, v) for k, v in info.items()
                              if k.startswith("_")])
            sub_infos = dict([(k, v) for k, v in info.items()
                              if not k.startswith("_")])
            if sub_infos:
                for sub, sub_info in sub_infos.items():
                    table.add_row([
                        "%s.%s" % (src, sub), "-", main_info["_version"], "-",
                        format_number(sub_info["_count"])
                    ])  # only _count avail there
            else:
                main_count = main_info.get("_count") and format_number(
                    main_info["_count"]) or ""
                table.add_row(
                    [src, "-",
                     main_info.get("_version", ""), "-", main_count])
        for src, info in sorted(self.changes["sources"]["deleted"].items(),
                                key=lambda e: e[0]):
            main_info = dict([(k, v) for k, v in info.items()
                              if k.startswith("_")])
            sub_infos = dict([(k, v) for k, v in info.items()
                              if not k.startswith("_")])
            if sub_infos:
                for sub, sub_info in sub_infos.items():
                    table.add_row([
                        "%s.%s" % (src, sub),
                        main_info.get("_version", ""), "-",
                        format_number(sub_info["_count"]), "-"
                    ])  # only _count avail there
            else:
                main_count = main_info.get("_count") and format_number(
                    main_info["_count"]) or ""
                table.add_row(
                    [src,
                     main_info.get("_version", ""), "-", main_count, "-"])
        for src, info in sorted(self.changes["sources"]["updated"].items(),
                                key=lambda e: e[0]):
            # extract information from main-source
            old_main_info = dict([(k, v) for k, v in info["old"].items()
                                  if k.startswith("_")])
            new_main_info = dict([(k, v) for k, v in info["new"].items()
                                  if k.startswith("_")])
            old_main_count = old_main_info.get("_count") and format_number(
                old_main_info["_count"]) or None
            new_main_count = new_main_info.get("_count") and format_number(
                new_main_info["_count"]) or None
            if old_main_count is None:
                assert new_main_count is None, "Sub-sources found for '%s', old and new count should " % src + \
                                               "both be None. Info was: %s" % info
                old_sub_infos = dict([(k, v) for k, v in info["old"].items()
                                      if not k.startswith("_")])
                new_sub_infos = dict([(k, v) for k, v in info["new"].items()
                                      if not k.startswith("_")])
                # old & new sub_infos should have the same structure (same existing keys)
                # so we just use one of them to explore
                if old_sub_infos:
                    assert new_sub_infos
                    for sub, sub_info in old_sub_infos.items():
                        table.add_row([
                            "%s.%s" % (src, sub),
                            old_main_info.get("_version", ""),
                            new_main_info.get("_version", ""),
                            format_number(sub_info["_count"]),
                            format_number(new_sub_infos[sub]["_count"])
                        ])
            else:
                assert new_main_count is not None, "No sub-sources found, old and new count should NOT " + \
                                                   "both be None. Info was: %s" % info
                table.add_row([
                    src,
                    old_main_info.get("_version", ""),
                    new_main_info.get("_version", ""), old_main_count,
                    new_main_count
                ])

        if table._rows:
            txt += table.get_string()
            txt += "\n"
        else:
            txt += "No datasource changed.\n"

        total_count = self.changes["new"].get("_count")
        if self.changes["sources"]["added"]:
            txt += "New datasource(s): %s\n" % ", ".join(
                sorted(list(self.changes["sources"]["added"])))
        if self.changes["sources"]["deleted"]:
            txt += "Deleted datasource(s): %s\n" % ", ".join(
                sorted(list(self.changes["sources"]["deleted"])))
        if self.changes["sources"]:
            txt += "\n"

        table = prettytable.PrettyTable(["Updated stats.", "previous", "new"])
        table.align["Updated stats."] = "l"
        table.align["previous"] = "r"
        table.align["new"] = "r"
        for stat_name, stat in sorted(self.changes["stats"]["added"].items(),
                                      key=lambda e: e[0]):
            table.add_row([stat_name, "-", format_number(stat["_count"])])
        for stat_name, stat in sorted(self.changes["stats"]["deleted"].items(),
                                      key=lambda e: e[0]):
            table.add_row([stat_name, format_number(stat["_count"]), "-"])
        for stat_name, stat in sorted(self.changes["stats"]["updated"].items(),
                                      key=lambda e: e[0]):
            table.add_row([
                stat_name,
                format_number(stat["old"]["_count"]),
                format_number(stat["new"]["_count"])
            ])
        if table._rows:
            txt += table.get_string()
            txt += "\n\n"

        if self.changes["new"]["_fields"]:
            new_fields = sorted(self.changes["new"]["_fields"].get("add", []))
            deleted_fields = self.changes["new"]["_fields"].get("remove", [])
            updated_fields = self.changes["new"]["_fields"].get("replace", [])
            if new_fields:
                txt += "New field(s): %s\n" % ", ".join(new_fields)
            if deleted_fields:
                txt += "Deleted field(s): %s\n" % ", ".join(deleted_fields)
            if updated_fields:
                txt += "Updated field(s): %s\n" % ", ".join(updated_fields)
            txt += "\n"

        if total_count is not None:
            txt += "Overall, %s documents in this release\n" % (
                format_number(total_count))
        if self.changes["new"]["_summary"]:
            sumups = []
            sumups.append(
                "%s document(s) added" %
                format_number(self.changes["new"]["_summary"].get("add", 0)))
            sumups.append("%s document(s) deleted" % format_number(
                self.changes["new"]["_summary"].get("delete", 0)))
            sumups.append("%s document(s) updated" % format_number(
                self.changes["new"]["_summary"].get("update", 0)))
            txt += ", ".join(sumups) + "\n"
        else:
            txt += "No information available for added/deleted/updated documents\n"

        if self.changes.get("note"):
            txt += "\n"
            txt += "Note: %s\n" % self.changes["note"]

        with open(filepath, "w") as fout:
            fout.write(txt)

        return txt



import sys
import re
import math
import os
import json
import time
import copy
import pickle
from datetime import datetime
from pprint import pformat
import asyncio
import aiocron
from functools import partial

from .mapper import TransparentMapper
from ..dataload.uploader import ResourceNotReady
from .differ import set_pending_to_diff
from ..databuild.backend import SourceDocMongoBackend, TargetDocMongoBackend, \
    LinkTargetDocMongoBackend
from biothings.utils.common import timesofar, iter_n, dotdict, open_compressed_file, \
    find_classes_subclassing, get_random_string
from biothings.utils.mongo import doc_feeder, id_feeder
from biothings.utils.loggers import get_logger
from biothings.utils.manager import BaseManager
from biothings.utils.dataload import merge_struct
import biothings.utils.mongo as mongo
from biothings.utils.hub_db import get_source_fullname, get_src_build_config, \
    get_src_build, get_src_dump, get_src_master
from biothings import config as btconfig
from biothings.hub import UPLOADER_CATEGORY, BUILDER_CATEGORY
from .backend import create_backend

logging = btconfig.logger


class BuilderException(Exception):
    pass


class ResumeException(Exception):
    pass


class DataBuilder(object):
    """
    Generic data builder.
    """

    keep_archive = 10  # number of archived collection to keep. Oldest get dropped first.

    def __init__(self,
                 build_name,
                 source_backend,
                 target_backend,
                 log_folder,
                 doc_root_key="root",
                 mappers=[],
                 default_mapper_class=TransparentMapper,
                 sources=None,
                 target_name=None,
                 **kwargs):
        self.init_state()
        self.build_name = build_name
        self.sources = sources
        self.target_name = target_name
        self._partial_source_backend = None
        self._partial_target_backend = None
        if type(source_backend) == partial:
            self._partial_source_backend = source_backend
        else:
            self._state["source_backend"] = source_backend
        if type(target_backend) == partial:
            self._partial_target_backend = target_backend
        else:
            self._state["target_backend"] = target_backend
        # doc_root_key is a key name within src_build_config doc.
        # it's a list of datasources that are able to create a document
        # even it doesn't exist. It root documents list is not empty, then
        # any other datasources not listed there won't be able to create
        # a document, only will they able to update it.
        # If no root documets, any datasources can create/update a doc
        # and thus there's no priority nor limitations
        # note: negations can be used, like "!source1". meaning source1 is not
        # root document datasource.
        # Usefull to express; "all resources except source1"
        self.doc_root_key = doc_root_key
        # overall merge start time
        self.t0 = time.time()
        # step merge start time
        self.ti = time.time()
        self.logfile = None
        self.log_folder = log_folder
        self.mappers = {}
        self.timestamp = datetime.now()
        self.merge_stats = {}  # keep track of cnt per source, etc...
        self.src_meta = {
        }  # sources involved in this build (includes versions)
        self.stats = {}  # can be customized
        self.mapping = {}  # ES mapping (merged from src_master's docs)

        for mapper in mappers + [default_mapper_class()]:
            self.mappers[mapper.name] = mapper

        self.step = kwargs.get("step", 10000)
        self.prepared = False

    def init_state(self):
        self._state = {
            "logger": None,
            "source_backend": None,
            "target_backend": None,
            "build_config": None,
        }

    @property
    def logger(self):
        if not self._state["logger"]:
            self.prepare()
        return self._state["logger"]

    @property
    def source_backend(self):
        if self._state["source_backend"] is None:
            self.prepare()
            self._state["build_config"] = self._state[
                "source_backend"].get_build_configuration(self.build_name)
            self._state["source_backend"].validate_sources(self.sources)
        return self._state["source_backend"]

    @property
    def target_backend(self):
        if self._state["target_backend"] is None:
            self.prepare()
        return self._state["target_backend"]

    @property
    def build_config(self):
        self._state[
            "build_config"] = self.source_backend.get_build_configuration(
                self.build_name)
        return self._state["build_config"]

    @logger.setter
    def logger(self, value):
        self._state["logger"] = value

    @build_config.setter
    def build_config(self, value):
        self._state["build_config"] = value

    def prepare(self, state={}):
        if self.prepared:
            return
        if state:
            # let's be explicit, _state takes what it wants
            for k in self._state:
                self._state[k] = state[k]
            return
        if self._partial_source_backend:
            self._state["source_backend"] = self._partial_source_backend()
        if self._partial_target_backend:
            self._state["target_backend"] = self._partial_target_backend()
        self.setup()
        self.setup_log()
        self.prepared = True

    def unprepare(self):
        """
        reset anything that's not pickable (so self can be pickled)
        return what's been reset as a dict, so self can be restored
        once pickled
        """
        # TODO: use copy ?
        state = {
            "logger": self._state["logger"],
            "source_backend": self._state["source_backend"],
            "target_backend": self._state["target_backend"],
            "build_config": self._state["build_config"],
        }
        for k in state:
            self._state[k] = None
        self.prepared = False
        return state

    def get_predicates(self):
        """
        Return a list of predicates (functions returning true/false, as in math logic)
        which instructs/dictates if job manager should start a job (process/thread)
        """
        def no_uploader_running(job_manager):
            """Uploaders could change the data to be merged..."""
            return len([
                j for j in job_manager.jobs.values()
                if j["category"] == UPLOADER_CATEGORY
            ]) == 0

        #def no_merger_running():
        #    """
        #    Mergers use cache files, if more than one running and caches need to be built
        #    both would try to write on the same cache file
        #    """
        #    return len([j for j in job_manager.jobs.values() if j["category"] == BUILDER_CATEGORY]) == 0
        return [no_uploader_running]

    def get_pinfo(self):
        """
        Return dict containing information about the current process
        (used to report in the hub)
        """
        pinfo = {
            "category": BUILDER_CATEGORY,
            "source":
            "%s:%s" % (self.build_name, self.target_backend.target_name),
            "step": "",
            "description": ""
        }
        preds = self.get_predicates()
        if preds:
            pinfo["__predicates__"] = preds
        return pinfo

    def setup_log(self):
        self.logger, _ = get_logger('build_%s' % self.build_name)

    def check_ready(self, force=False):
        if force:
            # don't even bother
            return
        src_build_config = self.source_backend.build_config
        src_dump = self.source_backend.dump
        _cfg = src_build_config.find_one({'_id': self.build_config['name']})
        # check if all resources are uploaded
        for src_name in _cfg["sources"]:
            fullname = get_source_fullname(src_name)
            if not fullname:
                raise ResourceNotReady("Can't find source '%s'" % src_name)
            main_name = fullname.split(".")[0]
            src_doc = src_dump.find_one({"_id": main_name})
            if not src_doc:
                raise ResourceNotReady(
                    "Missing information for source '%s' to start merging" %
                    src_name)
            if not src_doc.get("upload", {}).get("jobs", {}).get(
                    src_name, {}).get("status") == "success":
                raise ResourceNotReady(
                    "No successful upload found for resource '%s'" % src_name)

    def get_target_name(self):
        return '{}_{}_{}'.format(self.build_name,
                                 self.get_build_version(),
                                 get_random_string()).lower()

    def get_build_version(self):
        """
        Generate an arbitrary major build version. Default is using a timestamp (YYMMDD)
        '.' char isn't allowed in build version as it's reserved for minor versions
        """
        d = datetime.fromtimestamp(self.t0)
        version_fmt = self.build_config.get("build_version")
        if not version_fmt:
            version_fmt = "%Y%m%d"
        bversion = d.strftime(version_fmt)
        self.logger.info("Build version: %s", bversion)

        return bversion

    def register_status(self, status, transient=False, init=False, **extra):
        """
        Register current build status. A build status is a record in src_build
        The key used in this dict the target_name. Then, any operation
        acting on this target_name is registered in a "jobs" list.
        """
        assert self.build_config, "build_config needs to be specified first"
        # get it from source_backend, kind of weird...
        src_build = self.source_backend.build
        all_sources = self.build_config.get("sources", [])
        target_name = "%s" % self.target_backend.target_name
        backend_url = self.target_backend.get_backend_url()
        build_info = {
            '_id': target_name,
            # TODO: deprecate target_backend & target_name, use backed_url instead
            'target_backend': self.target_backend.name,
            'target_name': target_name,
            'backend_url': backend_url,
            'build_config': self.build_config,
            # these are all the sources required to build target
            # (not just the ones being processed, those are registered in jobs
            'sources': all_sources,
        }
        job_info = {
            'status': status,
            'step_started_at': datetime.now(),
            'logfile': self.logfile,
        }
        if transient:
            # record some "in-progress" information
            job_info['pid'] = os.getpid()
        else:
            # only register time when it's a final state
            job_info["time"] = timesofar(self.ti)
            t1 = round(time.time() - self.ti, 0)
            job_info["time_in_s"] = t1
        if "build" in extra:
            build_info.update(extra["build"])
        if "job" in extra:
            job_info.update(extra["job"])
        # create a new build entry in "build" dict if none exists
        build = src_build.find_one({'_id': target_name})
        if not build:
            # first record for target_name, keep a timestamp
            build_info["started_at"] = datetime.fromtimestamp(self.t0)
            build_info["jobs"] = []
            src_build.insert_one(build_info)
        if init:
            # init timer for this step
            self.ti = time.time()
            src_build.update({'_id': target_name},
                             {"$push": {
                                 'jobs': job_info
                             }})
            # now refresh/sync
            build = src_build.find_one({'_id': target_name})
        else:
            # merge extra at root level
            # (to keep building data...) and update the last one
            # (it's been properly created before when init=True)
            build["jobs"] and build["jobs"][-1].update(job_info)

            # build_info is common to all jobs, so we want to keep
            # any existing data (well... except if it's explicitely specified)
            def merge_build_info(target, d):
                if "__REPLACE__" in d.keys():
                    d.pop("__REPLACE__")
                    target = d
                else:
                    for k, v in d.items():
                        if type(v) == dict:
                            if k in target:
                                target[k] = merge_build_info(target[k], v)
                            else:
                                v.pop("__REPLACE__", None)
                                # merge v with "nothing" just to make sure to remove any "__REPLACE__"
                                v = merge_build_info({}, v)
                                target[k] = v
                        else:
                            target[k] = v
                return target

            build = merge_build_info(build, build_info)
            src_build.replace_one({"_id": build["_id"]}, build)

    def clean_old_collections(self):
        # use target_name is given, otherwise build name will be used
        # as collection name prefix, so they should start like that
        prefix = "%s_" % (self.target_name or self.build_name)
        db = mongo.get_target_db()
        cols = [c for c in db.collection_names() if c.startswith(prefix)]
        # timestamp is what's after _archive_, YYYYMMDD, so we can sort it safely
        cols = sorted(cols, reverse=True)
        to_drop = cols[self.keep_archive:]
        for colname in to_drop:
            self.logger.info("Cleaning old archive collection '%s'" % colname)
            db[colname].drop()

    def init_mapper(self, mapper_name):
        if self.mappers[mapper_name].need_load():
            if mapper_name is None:
                self.logger.info("Initializing default mapper")
            else:
                self.logger.info("Initializing mapper name '%s'" % mapper_name)
            self.mappers[mapper_name].load()

    def generate_document_query(self, src_name):
        return None

    def get_root_document_sources(self):
        root_srcs = self.build_config.get(self.doc_root_key, []) or []
        # check for "not this resource" and adjust the list
        none_root_srcs = [
            src.replace("!", "") for src in root_srcs if src.startswith("!")
        ]
        if none_root_srcs:
            if len(none_root_srcs) != len(root_srcs):
                raise BuilderException("If using '!' operator, all datasources must use it (cannot mix), got: %s" %
                                       (repr(root_srcs)))
            # ok, grab sources for this build,
            srcs = self.build_config.get("sources", [])
            root_srcs = list(set(srcs).difference(set(none_root_srcs)))
            #self.logger.info("'except root' sources %s resolves to root source = %s" % (repr(none_root_srcs),root_srcs))

        # resolve possible regex based source name (split-collections sources)
        root_srcs = self.resolve_sources(root_srcs)
        return root_srcs

    def setup(self, sources=None, target_name=None):
        sources = sources or self.sources
        # try to get target_name from args, otherwise for now generate it
        # using mongo backend (it'll be set later during merge() call)
        target_name = target_name or self.target_name
        self.target_backend.set_target_name(self.target_name, self.build_name)
        # root key is optional but if set, it must exist in build config
        if self.doc_root_key and self.doc_root_key not in self.build_config:
            raise BuilderException(
                "Root document key '%s' can't be found in build configuration"
                % self.doc_root_key)

    def get_stats(self, sources, job_manager):
        """
        Return a dictionnary of metadata for this build. It's usually app-specific
        and this method may be overridden as needed. By default though, the total
        number of documents in the merged collection is stored (key "total")

        Return dictionary will be merged with any existing metadata in
        src_build collection. This behavior can be changed by setting a special
        key within metadata dict: {"__REPLACE__" : True} will... replace existing
        metadata with the one returned here.

        "job_manager" is passed in case parallelization is needed. Be aware
        that this method is already running in a dedicated thread, in order to
        use job_manager, the following code must be used at the very beginning
        of its implementation:
        asyncio.set_event_loop(job_manager.loop)
        """
        total = self.target_backend.target_collection.count()
        return {"total": total}

    def get_custom_metadata(self, sources, job_manager):
        """
        If more metadata is required, this method can be overridden and should
        return a dict. Existing metadata dict will be update with that one
        before storage.
        """
        return {}

    def get_mapping(self, sources):
        """
        Merge mappings from src_master
        """
        mapping = {}
        src_master = self.source_backend.master
        for collection in self.build_config['sources']:
            meta = src_master.find_one({"_id": collection})
            if 'mapping' in meta and meta["mapping"]:
                mapping = merge_struct(mapping, meta['mapping'])
            else:
                raise BuilderException('"%s" has no mapping data' % collection)
        return mapping

    def store_metadata(self, res, sources, job_manager):
        self.target_backend.post_merge()
        self.src_meta = self.source_backend.get_src_metadata()
        # now that we have merge stats (count/srcs) + all src involved
        # we can propagate stats
        self.update_src_meta_stats()
        self.mapping = self.get_mapping(sources)
        self.stats = self.get_stats(sources, job_manager)
        self.custom_metadata = self.get_custom_metadata(sources, job_manager)
        # also search for _meta in build_config
        bmeta = self.build_config.get("_meta")
        if bmeta:
            self.logger.info("Found _meta in build_config, merging: %s" %
                             pformat(bmeta))
            self.custom_metadata.update(self.build_config.get("_meta", {}))

    def update_src_meta_stats(self):
        for src, count in self.merge_stats.items():
            mainsrc = get_source_fullname(src).split(".")[0]
            self.src_meta.setdefault(mainsrc, {}).setdefault("stats", {})
            self.src_meta[mainsrc]["stats"].update({src: count})

    def resolve_sources(self, sources):
        """
        Source can be a string that may contain regex chars. It's usefull
        when you have plenty of sub-collections prefixed with a source name.
        For instance, given a source named "blah" stored in as many collections
        as chromosomes, insteand of passing each name as "blah_1", "blah_2", etc...
        "blah_.*" can be specified in build_config. This method resolves potential
        regexed source name into real, existing collection names
        """
        if type(sources) == str:
            sources = [sources]
        src_db = mongo.get_src_db()
        cols = src_db.collection_names()
        masters = self.source_backend.get_src_master_docs()
        found = []
        for src in sources:
            # check if master _id and name are different (meaning name is a regex)
            master = masters.get(src)
            if not master:
                raise BuilderException("'%s'could not be found in master documents (%s)" %
                                       (src, repr(list(masters.keys()))))
            search = src
            if master["_id"] != master["name"]:
                search = master["name"]
            # restrict pattern to minimal match
            pat = re.compile("^%s$" % search)
            for col in cols:
                if pat.match(col):
                    found.append(col)
        return found

    def merge(self,
              sources=None,
              target_name=None,
              force=False,
              ids=None,
              steps=["merge", "post", "metadata"],
              job_manager=None,
              *args,
              **kwargs):
        """Merge given sources into a collection named target_name. If sources argument is omitted,
        all sources defined for this merger will be merged together, according to what is defined
        insrc_build_config. If target_name is not defined, a unique name will be generated.

        Optional parameters:
          - force=True will bypass any safety check
          - ids: list of _ids to merge, specifically. If None, all documents are merged.
          - steps:
             * merge: actual merge step, create merged documents and store them
             * post: once merge, run optional post-merge process
             * metadata: generate and store metadata (depends on merger, usually specifies the amount
                         of merged data, source versions, etc...)
        """
        assert job_manager
        # check what to do
        if type(steps) == str:
            steps = [steps]
        self.t0 = time.time()
        self.check_ready(force)
        # normalize
        avail_sources = self.build_config['sources']
        if sources is None:
            self.target_backend.drop()
            self.target_backend.prepare()
            sources = avail_sources  # merge all
        elif isinstance(sources, str):
            sources = [sources]

        if ids is None:
            # nothing passed specifically, let's have a look at the config
            ids = self.build_config.get("ids")
            if ids:
                # config calls for a merge on specific _ids
                if type(ids) == str:
                    # path to a file
                    m = map(lambda l: l.decode().strip(),
                            open_compressed_file(ids).readlines())
                    ids = [_id for _id in m if not _id.startswith("#")]

        orig_sources = sources
        sources = self.resolve_sources(sources)
        if not sources and "merge" in steps:
            raise BuilderException("No source found, got %s while available sources are: %s" %
                                   (repr(orig_sources), repr(avail_sources)))
        if not target_name:
            target_name = self.get_target_name()
        self.target_name = target_name
        self.target_backend.set_target_name(self.target_name)

        self.custom_metadata = {}
        self.clean_old_collections()
        self.logger.info("Merging into target collection '%s'" %
                         self.target_backend.target_name)
        strargs = "[sources=%s,target_name=%s]" % (sources, target_name)

        try:

            @asyncio.coroutine
            def do():
                res = None
                if "merge" in steps or "post" in steps:
                    job = self.merge_sources(source_names=sources,
                                             ids=ids,
                                             steps=steps,
                                             job_manager=job_manager,
                                             *args,
                                             **kwargs)
                    res = yield from job
                if "metadata" in steps:
                    pinfo = self.get_pinfo()
                    pinfo["step"] = "metadata"
                    self.register_status("building",
                                         transient=True,
                                         init=True,
                                         job={"step": "metadata"})
                    postjob = yield from job_manager.defer_to_thread(
                        pinfo,
                        partial(self.store_metadata,
                                res,
                                sources=sources,
                                job_manager=job_manager))

                    def stored(f):
                        try:
                            nonlocal res
                            if res:
                                res = f.result(
                                )  # consume to trigger exceptions if any
                            strargs = "[sources=%s,stats=%s]" % \
                                (sources, self.merge_stats)
                            build_version = self.get_build_version()
                            if "." in build_version:
                                raise BuilderException(
                                    "Can't use '.' in build version '%s', it's reserved for minor versions"
                                    % build_version)
                            # get original start dt
                            src_build = self.source_backend.build
                            build = src_build.find_one({'_id': target_name})
                            _meta = {
                                "biothing_type":
                                build["build_config"]["doc_type"],
                                "src":
                                self.src_meta,
                                "stats":
                                self.stats,
                                "build_version":
                                build_version,
                                "build_date":
                                datetime.fromtimestamp(self.t0).isoformat()
                            }
                            # custom
                            _meta.update(self.custom_metadata)
                            self.register_status('success',
                                                 build={
                                                     "merge_stats":
                                                     self.merge_stats,
                                                     "mapping": self.mapping,
                                                     "_meta": _meta,
                                                 })
                            self.logger.info("success %s" % strargs,
                                             extra={"notify": True})
                            set_pending_to_diff(target_name)
                        except Exception as e:
                            strargs = "[sources=%s]" % sources
                            self.register_status("failed",
                                                 job={"err": repr(e)})
                            self.logger.exception("failed %s: %s" %
                                                  (strargs, e),
                                                  extra={"notify": True})
                            raise

                    postjob.add_done_callback(stored)
                    yield from postjob

            task = asyncio.ensure_future(do())
            return task

        except (KeyboardInterrupt, Exception) as e:
            self.logger.exception(e)
            self.register_status("failed", job={"err": repr(e)})
            self.logger.exception("failed %s: %s" % (strargs, e),
                                  extra={"notify": True})
            raise

    def get_mapper_for_source(self, src_name, init=True):
        # src_name can be a regex (when source has split collections, they are merge but
        # comes from the same "template" sourcek
        docs = self.source_backend.get_src_master_docs()
        mapper_name = None
        for master_name in docs:
            pat = re.compile("^%s$" % master_name)
            if pat.match(src_name):
                mapper_name = docs[master_name].get("mapper")
        # TODO: this could be a list
        try:
            init and self.init_mapper(mapper_name)
            mapper = self.mappers[mapper_name]
            self.logger.info("Found mapper '%s' for source '%s'" %
                             (mapper, src_name))
            return mapper
        except KeyError:
            raise BuilderException(
                "Found mapper named '%s' but no mapper associated" %
                mapper_name)

    @asyncio.coroutine
    def merge_sources(self,
                      source_names,
                      steps=["merge", "post"],
                      batch_size=100000,
                      ids=None,
                      job_manager=None):
        """
        Merge resources from given source_names or from build config.
        Identify root document sources from the list to first process them.
        ids can a be list of documents to be merged in particular.
        """
        assert job_manager
        # check what to do
        if type(steps) == str:
            steps = [steps]
        do_merge = "merge" in steps
        do_post_merge = "post" in steps
        self.merge_stats = {}
        self.stats = {}
        self.mapping = {}
        # try to identify root document sources amongst the list to first
        # process them (if any)
        defined_root_sources = self.get_root_document_sources()
        root_sources = list(
            set(source_names).intersection(set(defined_root_sources)))
        other_sources = list(set(source_names).difference(set(root_sources)))
        # got root doc sources but not part of the merge ? that's weird...
        if defined_root_sources and not root_sources:
            self.logger.warning(
                "Root document sources found (%s) but not part of the merge..."
                % defined_root_sources)

        source_names = sorted(source_names)
        root_sources = sorted(root_sources)
        other_sources = sorted(other_sources)

        self.logger.info("Sources to be merged: %s" % source_names)
        self.logger.info("Root sources: %s" % root_sources)
        self.logger.info("Other sources: %s" % other_sources)

        got_error = False

        @asyncio.coroutine
        def merge(src_names):
            jobs = []
            for i, src_name in enumerate(src_names):
                yield from asyncio.sleep(0.0)
                job = self.merge_source(src_name,
                                        batch_size=batch_size,
                                        ids=ids,
                                        job_manager=job_manager)
                job = asyncio.ensure_future(job)

                def merged(f, name, stats):
                    try:
                        res = f.result()
                        stats.update(res)
                    except Exception as e:
                        self.logger.exception(
                            "Failed merging source '%s': %s" % (name, e))
                        nonlocal got_error
                        got_error = e

                job.add_done_callback(
                    partial(merged, name=src_name, stats=self.merge_stats))
                jobs.append(job)
                yield from asyncio.wait([job])
                # raise error as soon as we know something went wrong
                if got_error:
                    raise got_error
            tasks = asyncio.gather(*jobs)
            yield from tasks

        if do_merge:
            if root_sources:
                self.register_status("building",
                                     transient=True,
                                     init=True,
                                     job={
                                         "step": "merge-root",
                                         "sources": root_sources
                                     })
                self.logger.info("Merging root document sources: %s" %
                                 root_sources)
                yield from merge(root_sources)
                self.register_status("success",
                                     job={
                                         "step": "merge-root",
                                         "sources": root_sources
                                     })

            if other_sources:
                self.register_status("building",
                                     transient=True,
                                     init=True,
                                     job={
                                         "step": "merge-others",
                                         "sources": other_sources
                                     })
                self.logger.info("Merging other resources: %s" % other_sources)
                yield from merge(other_sources)
                self.register_status("success",
                                     job={
                                         "step": "merge-others",
                                         "sources": other_sources
                                     })

            self.register_status("building",
                                 transient=True,
                                 init=True,
                                 job={"step": "finalizing"})
            self.logger.info("Finalizing target backend")
            self.target_backend.finalize()
            self.register_status("success", job={"step": "finalizing"})
        else:
            self.logger.info("Skip data merging")

        if do_post_merge:
            self.logger.info("Running post-merge process")
            self.register_status("building",
                                 transient=True,
                                 init=True,
                                 job={"step": "post-merge"})
            pinfo = self.get_pinfo()
            pinfo["step"] = "post-merge"
            job = yield from job_manager.defer_to_thread(
                pinfo,
                partial(self.post_merge, source_names, batch_size,
                        job_manager))
            job = asyncio.ensure_future(job)

            def postmerged(f):
                try:
                    self.logger.info("Post-merge completed [%s]" % f.result())
                    self.register_status("success", job={"step": "post-merge"})
                except Exception as e:
                    self.logger.exception("Failed post-merging source: %s" % e)
                    nonlocal got_error
                    got_error = e

            job.add_done_callback(postmerged)
            yield from job
            if got_error:
                raise got_error
        else:
            self.logger.info("Skip post-merge process")

        yield from asyncio.sleep(0.0)
        return self.merge_stats

    def document_cleaner(self, src_name, *args, **kwargs):
        """
        Return a function taking a document as argument, cleaning the doc
        as needed, and returning that doc. If no function is needed, None.
        Note: the returned function must be pickleable, careful with lambdas
        and closures.
        """
        return None

    @asyncio.coroutine
    def merge_source(self,
                     src_name,
                     batch_size=100000,
                     ids=None,
                     job_manager=None):
        # it's actually not optional
        assert job_manager
        _query = self.generate_document_query(src_name)
        # Note: no need to check if there's an existing document with _id (we want to merge only with an existing document)
        # if the document doesn't exist then the update() call will silently fail.
        # That being said... if no root documents, then there won't be any previously inserted
        # documents, and this update() would just do nothing. So if no root docs, then upsert
        # (update or insert, but do something)
        defined_root_sources = self.get_root_document_sources()
        upsert = not defined_root_sources or src_name in defined_root_sources
        if not upsert:
            self.logger.debug(
                "Documents from source '%s' will be stored only if a previous document exists with same _id"
                % src_name)
        jobs = []
        total = self.source_backend[src_name].count()
        btotal = math.ceil(total / batch_size)
        bnum = 1
        cnt = 0
        got_error = False
        # grab ids only, so we can get more, let's say 10 times more
        id_batch_size = batch_size * 10
        if ids:
            self.logger.info(
                "Merging '%s' specific list of _ids, create merger job with batch_size=%d"
                % (src_name, batch_size))
            id_provider = [ids]
        else:
            self.logger.info(
                "Fetch _ids from '%s' with batch_size=%d, and create merger job with batch_size=%d"
                % (src_name, id_batch_size, batch_size))
            id_provider = id_feeder(self.source_backend[src_name],
                                    batch_size=id_batch_size)

        if _query and ids is not None:
            self.logger.info(
                "Query/filter involved, but also specific list of _ids. Ignoring query and use _ids"
            )

        if _query and ids is None:
            self.logger.info(
                "Query/filter involved, can't use cache to fetch _ids")
            # use doc_feeder but post-process doc to keep only the _id
            id_provider = map(
                lambda docs: [d["_id"] for d in docs],
                doc_feeder(self.source_backend[src_name],
                           query=_query,
                           step=batch_size,
                           inbatch=True,
                           fields={"_id": 1}))
        else:
            # when passing a list of _ids, IDs will be sent to the query, so we need to reduce the batch size
            id_provider = ids and iter_n(ids, int(
                batch_size / 100)) or id_feeder(self.source_backend[src_name],
                                                batch_size=id_batch_size,
                                                logger=self.logger)

        src_master = self.source_backend.master
        meta = src_master.find_one({"_id": src_name}) or {}
        merger = meta.get("merger", "upsert")
        self.logger.info("Documents from source '%s' will be merged using %s" %
                         (src_name, merger))

        doc_cleaner = self.document_cleaner(src_name)
        for big_doc_ids in id_provider:
            for doc_ids in iter_n(big_doc_ids, batch_size):
                # try to put some async here to give control back
                # (but everybody knows it's a blocking call: doc_feeder)
                yield from asyncio.sleep(0.1)
                cnt += len(doc_ids)
                pinfo = self.get_pinfo()
                pinfo["step"] = src_name
                pinfo["description"] = "#%d/%d (%.1f%%)" % (bnum, btotal,
                                                            (cnt / total * 100))
                self.logger.info("Creating merger job #%d/%d, to process '%s' %d/%d (%.1f%%)" %
                                 (bnum, btotal, src_name, cnt, total, (cnt/total*100.)))
                job = yield from job_manager.defer_to_process(
                    pinfo,
                    partial(merger_worker, self.source_backend[src_name].name,
                            self.target_backend.target_name, doc_ids,
                            self.get_mapper_for_source(src_name, init=False),
                            doc_cleaner, upsert, merger, bnum))

                def batch_merged(f, batch_num):
                    nonlocal got_error
                    if type(f.result()) != int:
                        got_error = Exception(
                            "Batch #%s failed while merging source '%s' [%s]" %
                            (batch_num, src_name, f.result()))

                job.add_done_callback(partial(batch_merged, batch_num=bnum))
                jobs.append(job)
                bnum += 1
                # raise error as soon as we know
                if got_error:
                    raise got_error
        self.logger.info("%d jobs created for merging step" % len(jobs))
        tasks = asyncio.gather(*jobs)

        def done(f):
            nonlocal got_error
            if None in f.result():
                got_error = Exception("Some batches failed")
                return
            # compute overall inserted/updated records (consume result() and check summable)
            _ = sum(f.result())

        tasks.add_done_callback(done)
        yield from tasks
        if got_error:
            raise got_error
        else:
            return {"%s" % src_name: cnt}

    def post_merge(self, source_names, batch_size, job_manager):
        pass


class LinkDataBuilder(DataBuilder):
    """
    LinkDataBuilder creates a link to the original datasource to be merged, without
    actually copying the data (merged collection remains empty). This builder is
    only valid when using only one datasource (thus no real merge) is declared in
    the list of sources to be merged, and is useful to prevent data duplication between
    the datasource itself and the resulting merged collection.
    """
    def __init__(self, build_name, source_backend, target_backend, *args,
                 **kwargs):

        super().__init__(build_name,
                         source_backend,
                         target_backend=partial(LinkTargetDocMongoBackend),
                         *args,
                         **kwargs)
        conf = self.source_backend.get_build_configuration(self.build_name)
        assert len(conf["sources"]) == 1, \
            "Found more than one source to link, not allowed: %s" % conf["sources"]
        assert hasattr(self.target_backend, "datasource_name")
        self.target_backend.datasource_name = conf["sources"][0]
        self.target_backend.source_db = self.source_backend

    @asyncio.coroutine
    def merge_source(self, src_name, *args, **kwargs):
        total = self.source_backend[src_name].count()
        return {"%s" % src_name: total}


from biothings.utils.backend import DocMongoBackend


def fix_batch_duplicates(docs, fail_if_struct_is_different=False):
    """
    Remove duplicates from docs based on _id. If _id's the same but
    structure is different (not real "duplicates", but different documents
    with the same _ids), merge docs all together (dict.update)
    or raise an error if fail_if_struct_is_different.
    """
    dids = {}
    # docs per _id
    for d in docs:
        dids.setdefault(d["_id"], []).append(d)
    # now check doc structure for each duplicates
    # if same structure, replace with one occurence of the docs
    # if not the same, log all the docs as warning, and merge them all
    # as we would do if we were upserting doc one-by-one (no batch)
    # note: dict are unhashable (no set) so either compare one each other (n^2-ish)
    # or use json strings (let's try json...)
    for _id in dids:
        jl = set([json.dumps(e, sort_keys=True) for e in dids[_id]])
        if len(jl) > 1:
            # different structure
            if fail_if_struct_is_different:
                raise ValueError(
                    "Found duplicated with different document structure: %s" %
                    dids[_id])
            else:
                logging.warning(
                    "Found duplicated with different document structure, merging them altogether: %s"
                    % dids[_id])
            # merge docs on top of each other
            dupdocs = dids[_id]
            merged = {}
            [merged.update(d) for d in dupdocs]
            dids[_id] = merged
        else:
            assert len(jl) == 1
            # normalize to scalar
            dids[_id] = dids[_id][0]

    return list(dids.values())


def merger_worker(col_name, dest_name, ids, mapper, cleaner, upsert, merger,
                  batch_num):
    try:
        src = mongo.get_src_db()
        tgt = mongo.get_target_db()
        col = src[col_name]
        dest = DocMongoBackend(tgt, tgt[dest_name])
        cur = doc_feeder(col,
                         step=len(ids),
                         inbatch=False,
                         query={'_id': {
                             '$in': ids
                         }})
        if cleaner:
            cur = map(cleaner, cur)
        mapper.load()
        docs = [d for d in mapper.process(cur)]
        # while documents from cursor "cur" are unique, at this point, due to the use
        # a mapper, documents can be converted and there now can be duplicates (same _id)
        # (ex: mygene, ensembl -> entrez conversion). "docs" could produce a duplicated error
        # within the batch, so we need to remove duplicates.
        all_ids = [d["_id"] for d in docs]
        uniq_ids = set(all_ids)
        if len(all_ids) != len(uniq_ids):
            logging.warning("Found duplicated IDs within batch, trying to fix")
            docs = fix_batch_duplicates(docs)
        if merger == "merge_struct":
            stored_docs = dest.mget_from_ids([d["_id"] for d in docs])
            ddocs = dict([(d["_id"], d) for d in docs])
            for d in stored_docs:
                ddocs[d["_id"]] = merge_struct(d, ddocs[d["_id"]])
            docs = list(ddocs.values())
        cnt = dest.update(docs, upsert=upsert)
        return cnt
    except Exception as e:
        logger_name = "build_%s_%s_batch_%s" % (dest_name, col_name, batch_num)
        logger, _ = get_logger(logger_name, btconfig.LOG_FOLDER)
        logger.exception(e)
        logger.error("col_name: %s, dest_name: %s, ids: see pickle, " % (col_name, dest_name)
                     + "mapper: %s, cleaner: %s, upsert: %s, " % (mapper, cleaner, upsert)
                     + "merger: %s, batch_num: %s" % (merger, batch_num))
        exc_fn = os.path.join(btconfig.LOG_FOLDER, "%s.exc.pick" % logger_name)
        pickle.dump(e, open(exc_fn, "wb"))
        logger.info("Exception was dumped in pickle file '%s'" % exc_fn)
        ids_fn = os.path.join(btconfig.LOG_FOLDER, "%s.ids.pick" % logger_name)
        pickle.dump(ids, open(ids_fn, "wb"))
        logger.info("IDs dumped in pickle file '%s'" % ids_fn)
        dat_fn = os.path.join(btconfig.LOG_FOLDER,
                              "%s.docs.pick" % logger_name)
        pickle.dump(docs, open(dat_fn, "wb"))
        logger.info("Data (batch of docs) dumped in pickle file '%s'" % dat_fn)
        raise


def set_pending_to_build(conf_name=None):
    src_build_config = get_src_build_config()
    qfilter = {}
    if conf_name:
        qfilter = {"_id": conf_name}
    logging.info("Setting pending_to_build flag for configuration(s): %s" %
                 (conf_name and conf_name or "all configuraitons"))
    src_build_config.update(qfilter, {"$addToSet": {"pending": "build"}})


class BuilderManager(BaseManager):
    def __init__(self,
                 source_backend_factory=None,
                 target_backend_factory=None,
                 builder_class=None,
                 poll_schedule=None,
                 *args,
                 **kwargs):
        """
        BuilderManager deals with the different builders used to merge datasources.
        It is connected to src_build() via sync(), where it grabs build information
        and register builder classes, ready to be instantiate when triggering builds.
        source_backend_factory can be a optional factory function (like a partial) that
        builder can call without any argument to generate a SourceBackend.
        Same for target_backend_factory for the TargetBackend. builder_class if given
        will be used as the actual Builder class used for the merge and will be passed
        same arguments as the base DataBuilder. It can also be a list of classes, in which
        case the default used one is the first, when it's necessary to define multiple builders.
        """
        super(BuilderManager, self).__init__(*args, **kwargs)
        self.src_build_config = get_src_build_config()
        self.source_backend_factory = source_backend_factory
        self.target_backend_factory = target_backend_factory
        builder_class = builder_class or DataBuilder
        if isinstance(builder_class, list):
            self.arg_builder_classes = builder_class
        else:
            self.arg_builder_classes = [builder_class]
        self.default_builder_class = self.arg_builder_classes[0] or DataBuilder
        self.builder_classes = {}
        self.poll_schedule = poll_schedule
        self.setup_log()

    def clean_stale_status(self):
        src_build = get_src_build()
        for build in src_build.find():
            dirty = False
            for job in build.get("jobs", []):
                if job.get("status") == "building":
                    logging.warning(
                        "Found stale build '%s', marking build status as 'canceled'"
                        % build["_id"])
                    job["status"] = "canceled"
                    dirty = True
            if dirty:
                src_build.replace_one({"_id": build["_id"]}, build)

    @property
    def source_backend(self):
        source_backend = self.source_backend_factory and self.source_backend_factory() or \
            partial(SourceDocMongoBackend,
                    build_config=partial(get_src_build_config),
                    build=partial(get_src_build),
                    master=partial(get_src_master),
                    dump=partial(get_src_dump),
                    sources=partial(mongo.get_src_db))
        return source_backend

    @property
    def target_backend(self):
        target_backend = self.target_backend_factory and self.target_backend_factory() or \
            partial(TargetDocMongoBackend,
                    target_db=partial(mongo.get_target_db))
        return target_backend

    def get_builder_class(self, build_config_name):
        """
        builder class can be specified different way (in order):
        1. within the build_config document (so, per configuration)
        2. or defined in the builder manager (so, per manager)
        3. or default to DataBuilder
        """
        builder_class = None
        conf = self.src_build_config.find_one({"_id": build_config_name})
        if conf.get("builder_class"):
            builder_class = self.builder_classes[
                conf["builder_class"]]["class"]
        elif self.default_builder_class:
            builder_class = self.default_builder_class
        else:
            builder_class = DataBuilder

        return builder_class

    def register_builder(self, build_name):
        # will use partial to postponse object creations and their db connection
        # as we don't want to keep connection alive for undetermined amount of time
        # declare source backend
        def create(build_name):
            # postpone config import so app had time to set it up
            # before actual call time
            from biothings import config
            # assemble the whole
            klass = self.get_builder_class(build_name)
            self.logger.info("Build config '%s' will use builder class %s",
                             build_name, klass)
            bdr = klass(build_name,
                        source_backend=self.source_backend,
                        target_backend=self.target_backend,
                        log_folder=config.LOG_FOLDER)

            return bdr

        self.register[build_name] = partial(create, build_name)

    def get_builder(self, col_name):
        doc = get_src_build().find_one({"_id": col_name})
        if not doc:
            raise BuilderException("No such build named '%s'" % repr(col_name))
        assert "build_config" in doc, "Expecting build_config information"
        klass = self.get_builder_class(doc["build_config"]["name"])
        bdr = klass(doc["build_config"]["name"],
                    source_backend=self.source_backend,
                    target_backend=self.target_backend,
                    log_folder=btconfig.LOG_FOLDER)
        # overwrite with existing values
        bdr.build_config = doc["build_config"]
        bdr.target_backend.set_target_name(col_name)

        return bdr

    def delete_merged_data(self, merge_name):
        target_db = mongo.get_target_db()
        col = target_db[merge_name]
        col.drop()

    def delete_merge(self, merge_name):
        """Delete merged collections and associated metadata"""
        db = get_src_build()
        meta = db.find_one({"_id": merge_name})
        if meta:
            db.remove({"_id": merge_name})
        else:
            self.logger.warning(
                "No metadata found for merged collection '%s'" % merge_name)
        self.delete_merged_data(merge_name)

    def archive_merge(self, merge_name):
        """Delete merged collections and associated metadata"""
        db = get_src_build()
        meta = db.find_one({"_id": merge_name})
        if meta:
            meta["archived"] = datetime.now()
            db.replace_one({"_id": merge_name}, meta)
        else:
            self.logger.warning(
                "No metadata found for merged collection '%s'" % merge_name)
        self.delete_merged_data(merge_name)

    def get_query_for_list_merge(self, only_archived):
        q = {"archived": {"$exists": 0}}
        if only_archived:
            q = {"archived": {"$exists": 1}}
        return q

    def list_merge(self, build_config=None, only_archived=False):
        q = self.get_query_for_list_merge(only_archived)
        docs = get_src_build().find(q)
        by_confs = {}
        for d in docs:
            by_confs.setdefault(
                d.get("build_config", {}).get("name", None),
                []).append(d["_id"])
        if build_config:
            return sorted(by_confs.get(build_config, []))
        else:
            for conf in by_confs:
                by_confs[conf] = sorted(by_confs[conf])
            return by_confs

    def setup_log(self):
        self.logger, self.logfile = get_logger("buildmanager")

    def __getitem__(self, build_name):
        """
        Return an instance of a builder for the build named 'build_name'
        Note: each call returns a different instance (factory call behind the scene...)
        """
        # we'll get a partial class but will return an instance
        pclass = BaseManager.__getitem__(self, build_name)
        return pclass()

    def configure(self):
        """Sync with src_build_config and register all build config"""
        self.register = {}
        self.builder_classes = {}
        for conf in self.src_build_config.find():
            self.register_builder(conf["_id"])
        self.find_builder_classes()

    def resolve_builder_class(self, klass):
        """
        Resolve class/partial definition to (obj,"type","mod.class")
        where names (class name, module, docstring, etc...) can
        directly be accessed whether it's a standard class or not
        """
        obj = klass
        if type(klass) == partial:
            assert type(klass.func) == type
            btype = "partial"
            obj = klass.func
        elif type(klass) == type:
            btype = "class"
        else:
            raise TypeError("Unknown type for builder %s" % repr(klass))
        modstr = obj.__module__
        classstr = obj.__name__
        classpathstr = "%s.%s" % (modstr, classstr)

        return (obj, btype, classpathstr)

    def find_builder_classes(self):
        """
        Find all available build class:
         1. classes passed during manager init (build_class)
            (that includes the default builder)
         2. all subclassing DataBuilder in:
           a. biothings.hub.databuilder.*
           b. hub.databuilder.* (app-specific)
        """
        bclasses = set(self.arg_builder_classes)
        mods = [sys.modules[__name__]]
        try:
            import hub.databuild as m
            mods.append(m)
        except ImportError:
            pass
        for klass in find_classes_subclassing(mods, DataBuilder):
            bclasses.add(klass)

        for klass in bclasses:
            try:
                obj, btype, classpathstr = self.resolve_builder_class(klass)
                helpstr = obj.__doc__ and " ".join(
                    map(str.strip, obj.__doc__.splitlines()))
                self.builder_classes[classpathstr] = {
                    "desc": helpstr,
                    "type": btype,
                    "class": klass,
                    "default": klass == self.default_builder_class,
                }
            except Exception as e:
                logging.exception(
                    "Can't extract information from builder class %s: %s" %
                    (repr(klass), e))

    def merge(self, build_name, sources=None, target_name=None, **kwargs):
        """
        Trigger a merge for build named 'build_name'. Optional list of sources can be
        passed (one single or a list). target_name is the target collection name used
        to store to merge data. If none, each call will generate a unique target_name.
        """
        try:
            bdr = self[build_name]
            job = bdr.merge(sources,
                            target_name,
                            job_manager=self.job_manager,
                            **kwargs)
            return job
        except KeyError:
            raise BuilderException("No such builder for '%s'" % build_name)
        except ResourceNotReady as e:
            raise BuilderException(
                "Some datasources aren't ready for the merge: %s" % e)

    def list_sources(self, build_name):
        """
        List all registered sources used to trigger a build named 'build_name'
        """
        info = self.src_build_config.find_one({"_id": build_name})
        return info and info["sources"] or []

    def whatsnew(self, build_name=None, old=None):
        """
        Return datasources which have changed since last time
        (last time is datasource information from metadata, either from
        given old src_build doc name, or the latest found if old=None)
        """
        dbbuild = get_src_build()
        dbdump = get_src_dump()

        def whatsnewcomparedto(build_name, old=None):
            if old is None:
                # TODO: this will get big... but needs to be generic
                # because we handle different hub db backends (or it needs to be a
                # specific helper func to be defined all backends
                builds = dbbuild.find({"build_config.name": build_name})
                builds = sorted(builds, key=lambda e: e["started_at"])
                if builds:
                    old = builds[-1]
                else:
                    raise BuilderException(
                        "Can't find a build associated to config '%s'" %
                        build_name)
            else:
                old = dbbuild.find_one({"_id": old})
            meta_srcs = old.get("_meta", {}).get("src", {})
            new = {
                "old_build": {
                    "name": old["_id"],
                    "built_at": old["started_at"]
                },
                "sources": {}
            }
            for src_name, data in meta_srcs.items():
                try:
                    srcd = dbdump.find_one({"_id": src_name})
                    if srcd and not srcd.get("download") and srcd.get(
                            "upload"):
                        # this is a collection only source, find all releases in sub sources, hopefully all are the same
                        rels = [
                            sub["release"]
                            for sub in srcd["upload"]["jobs"].values()
                        ]
                        srels = set(rels)
                        if len(srels) != 1:
                            raise ValueError(
                                "Found different releases in sub-sources, expected only one common: %s"
                                % repr(rels))
                        rel = rels[0]
                        if data.get("version") and rel != data["version"]:
                            new["sources"][src_name] = {
                                "old": {
                                    "version": data["version"]
                                },
                                "new": {
                                    "version": rel
                                }
                            }
                    elif srcd and srcd.get("download", {}).get(
                            "release"
                    ) and srcd["download"]["release"] != data["version"]:
                        new["sources"][src_name] = {
                            "old": {
                                "version": data["version"]
                            },
                            "new": {
                                "version": srcd["download"]["release"],
                                "downloaded_at":
                                srcd["download"].get("started_at")
                            }
                        }
                except Exception as e:
                    self.logger.warning(
                        "Can't check what's new for source '%s': %s" %
                        (src_name, e))
            return {build_name: new}

        if old is None and build_name is None:
            # do this for all build configs
            dbbuildconfig = get_src_build_config()
            configs = {}
            for d in dbbuildconfig.find({
                    "$or": [{
                    "archived": {
                    "$exists": 0
                    }
                    }, {
                    "archived": False
                    }]}):
                try:
                    news = whatsnewcomparedto(d["_id"])
                    if news[d["_id"]]["sources"]:
                        configs.update(news)
                except BuilderException:
                    continue
            return configs
        else:
            return whatsnewcomparedto(build_name, old)

    def clean_temp_collections(self, build_name, date=None, prefix=''):
        """
        Delete all target collections created from builder named
        "build_name" at given date (or any date is none given -- carefull...).
        Date is a string (YYYYMMDD or regex)
        Common collection name prefix can also be specified if needed.
        """
        target_db = mongo.get_target_db()
        for col_name in target_db.collection_names():
            search = prefix and prefix + "_" or ""
            search += build_name + '_'
            search += date and date + '_' or ''
            pat = re.compile(search)
            if pat.match(col_name) and 'current' not in col_name:
                logging.info("Dropping target collection '%s" % col_name)
                target_db[col_name].drop()

    def poll(self):
        """
        Check "whatsnew()" to idenfity builds which could be automatically built,
        if {"autobuild" : {...}} is part of the build configuration. "autobuild" contains
        a dict with "schedule" (aiocron/crontab format), so each build configuration can
        have a different polling schedule.
        """
        # don't use $exists in find(), not all hub backend implements that
        confs = get_src_build_config().find()
        autobuild_confs = [conf for conf in confs if "autobuild" in conf]

        @asyncio.coroutine
        def check_new(conf_name):
            new = self.whatsnew(conf_name)
            if new[conf_name]["sources"]:
                new_sources = [(name, info["new"].get("version")) for (name, info) in new[conf_name]["sources"].items()]
                logging.info("Build configuration '%s' can be launched (autobuid) because" % conf_name
                             + "some datasources are new: %s" % new_sources)
                self.merge(conf_name)
            else:
                logging.debug("Nothing new for build configuration '%s'" % conf_name)

        for conf in autobuild_confs:
            try:
                sch = conf["autobuild"].get("schedule")
                if not sch:
                    logging.error("Build configuration needs autobuild but 'schedule' isn't defined" % conf_name)
                    continue
                aiocron.crontab(
                    sch, func=partial(check_new, conf["_id"]),
                    start=True, loop=self.job_manager.loop
                )
            except Exception as e:
                logging.exception("Invalid autobuild information for build config '%s': %s" % (conf["_id"], e))

    def trigger_merge(self, doc):
        return self.merge(doc["_id"])

    def build_config_info(self):
        configs = {}
        err = None
        for name in self.register:
            try:
                builder = self[name]
            except Exception as e:
                conf = get_src_build_config().find_one({"_id": name})
                if conf:
                    builder = dotdict({"build_config": conf})
                else:
                    builder = None
                err = str(e)

            if not builder or issubclass(builder.target_backend.__class__, LinkTargetDocMongoBackend) or \
                    issubclass(builder.__class__, dict):  # fake builder obj
                target_db = None  # it's not a traditional target database, it's pointing to
                # somewhere else (TODO: maybe LinkTargetDocMongoBackend should
                # implement more methods to return info about that
            else:
                target_db = builder.target_backend.target_collection.database.client.address
            configs[name] = {
                "build_config": builder and builder.build_config,
                "archived": "archived" in (builder and builder.build_config
                                           or [])
            }
            if builder and builder.source_backend:
                configs[name]["source_backend"] = {
                    "type":
                    builder and builder.source_backend.__class__.__name__,
                    "source_db":
                    builder and builder.source_backend.sources.client.address,
                }
            if builder and builder.target_backend:
                configs[name]["target_backend"] = {
                    "type": builder
                    and builder.target_backend.__class__.__name__,
                    "target_db": target_db
                }
            if err:
                configs[name]["error"] = err
            if builder and builder.mappers:
                configs[name]["mapper"] = {}
                for mappername, mapper in builder.mappers.items():
                    configs[name]["mapper"][
                        mappername] = mapper.__class__.__name__
        res = {"build_configs": configs}
        # dict contains an actual class, non-serializable, so adjust:
        bclasses = copy.deepcopy(self.builder_classes)
        for k, v in bclasses.items():
            v.pop("class")
        res["builder_classes"] = bclasses
        return res

    def build_info(self,
                   id=None,
                   conf_name=None,
                   fields=None,
                   only_archived=False):
        """
        Return build information given an build _id, or all builds
        if _id is None. "fields" can be passed to select which fields
        to return or not (mongo notation for projections), if None
        return everything except:
         - "mapping" (too long)
        If id is None, more are filtered:
         - "sources" and some of "build_config"
        only_archived=True will return archived merges only
        """
        res = {}
        q = self.get_query_for_list_merge(only_archived)
        if id is not None:
            q = {"_id": id}
        else:
            fields = {}
            fields["mapping"] = 0
            fields["sources"] = 0
            fields["build_config.sources"] = 0
            fields["build_config.root"] = 0
        if conf_name is not None:
            q["build_config._id"] = conf_name
        builds = [b for b in get_src_build().find(q, fields)]
        res = [
            b for b in sorted(
                builds, key=lambda e: str(e["started_at"]), reverse=True)
        ]
        # set a global status (ie. latest job's status)
        # + get total #docs
        db = mongo.get_target_db()
        for b in res:
            jobs = b.get("jobs", [])
            b["status"] = "unknown"
            if jobs:
                b["status"] = jobs[-1]["status"]
            try:
                backend = create_backend(b["backend_url"])
                b["count"] = backend.count()
            except KeyError:
                b["count"] = db[b["_id"]].count()

        if id:
            if res:
                return res.pop()
            else:
                raise ValueError("No such build named '%s'" % id)
        else:
            return res

    def upsert_build_conf(self, name, doc_type, sources, roots, builder_class,
                          params, archived):
        col = get_src_build_config()
        builder_class = builder_class or self.resolve_builder_class(
            self.default_builder_class)[2]  # class path string
        doc = {
            "_id": name,
            "name": name,
            "doc_type": doc_type,
            "sources": sources,
            "root": roots,
            "builder_class": builder_class
        }
        if archived:
            doc["archived"] = True
        else:
            doc.pop("archived", None)
        doc.update(params)
        col.save(doc)
        self.configure()

    def create_build_configuration(self,
                                   name,
                                   doc_type,
                                   sources,
                                   roots=[],
                                   builder_class=None,
                                   params={},
                                   archived=False):
        col = get_src_build_config()
        # check conf doesn't exist yet
        if [d for d in col.find({"_id": name})]:
            raise ValueError("Configuration named '%s' already exists" % name)
        self.upsert_build_conf(name, doc_type, sources, roots, builder_class,
                               params, archived)

    def update_build_configuration(self,
                                   name,
                                   doc_type,
                                   sources,
                                   roots=[],
                                   builder_class=None,
                                   params={},
                                   archived=False):
        self.upsert_build_conf(name, doc_type, sources, roots, builder_class,
                               params, archived)

    def delete_build_configuration(self, name):
        col = get_src_build_config()
        col.remove({"_id": name})
        self.configure()

    def save_mapping(self, name, mapping=None, dest="build", mode="mapping"):
        logging.debug("Saving mapping for build '%s' destination='%s':\n%s" %
                      (name, dest, pformat(mapping)))
        src_build = get_src_build()
        m = src_build.find_one({"_id": name})
        assert m, "Can't find build document for '%s'" % name
        # either given a fully qualified source or just sub-source
        if dest == "build":
            m["mapping"] = mapping
            src_build.save(m)
        elif dest == "inspect":
            try:
                m["inspect"]["results"][mode] = mapping
                src_build.save(m)
            except KeyError as e:
                raise ValueError(
                    "Can't save mapping, document doesn't contain expected inspection data"
                    % e)
        else:
            raise ValueError("Unknow saving destination: %s" % repr(dest))
